/*
 * Prototype parser for smtlib format
 */

%{
  open Smtlib
%}

%token OPEN_PAR CLOSE_PAR
%token PLUS MINUS TIMES DIVIDE OPPOSITE
%token ITE
%token LT GT LEQ GEQ
%token EQUAL DISTINCT
%token AND OR XOR IFF IMPLIES NOT TT FF
%token IF_THEN_ELSE
%token LET FLET

%token SAT UNSAT UNKNOWN

%token LOGIC_ATTRIBUTE
%token ASSUMPTION_ATTRIBUTE
%token FORMULA_ATTRIBUTE
%token STATUS_ATTRIBUTE
%token EXTRASORTS_ATTRIBUTE
%token EXTRAPREDS_ATTRIBUTE
%token EXTRAFUNS_ATTRIBUTE
%token NOTES_ATTRIBUTE
%token OTHER_ATTRIBUTE

%token BENCHMARK
%token <string> IDENT
%token <string> NUMBER
%token <string> FVAR
%token <string> VAR
%token STRING
%token USER_DATA

%token LEXERROR

%start benchmark

%type <unit> benchmark		    


%% 

attribute: 
    LOGIC_ATTRIBUTE              { () }
  | ASSUMPTION_ATTRIBUTE         { () }
  | FORMULA_ATTRIBUTE            { () }
  | STATUS_ATTRIBUTE             { () }
  | EXTRASORTS_ATTRIBUTE         { () }
  | EXTRAPREDS_ATTRIBUTE         { () }
  | EXTRAFUNS_ATTRIBUTE          { () }
  | NOTES_ATTRIBUTE              { () }
  | OTHER_ATTRIBUTE              { () }
;

annotation:
    attribute                    { () }
  | attribute USER_DATA          { () }
;

other_annotation:
    OTHER_ATTRIBUTE              { () }
  | OTHER_ATTRIBUTE USER_DATA    { () }
;

nonempty_list_of_annotations:
    annotation                   { () }
  | nonempty_list_of_annotations annotation   { () }
; 

list_of_annotations:
    nonempty_list_of_annotations    { () }
  |                                 { () }
;



benchmark:
    OPEN_PAR BENCHMARK IDENT list_of_attributes CLOSE_PAR
       { set_benchmark_name $3 }
;

list_of_attributes:
    list_of_attributes benchmark_attribute    { () }
  | benchmark_attribute                       { () }
;

benchmark_attribute:
    LOGIC_ATTRIBUTE IDENT                 { set_benchmark_logic $2 }
  | ASSUMPTION_ATTRIBUTE formula          { push_benchmark_assumption $2 }
  | FORMULA_ATTRIBUTE formula             { set_benchmark_formula $2 }
  | STATUS_ATTRIBUTE status               { set_benchmark_status $2 }

  | EXTRASORTS_ATTRIBUTE OPEN_PAR nonempty_list_of_sorts CLOSE_PAR  { () }
  | EXTRAFUNS_ATTRIBUTE OPEN_PAR nonempty_list_of_funs CLOSE_PAR    { () }
  | EXTRAPREDS_ATTRIBUTE OPEN_PAR nonempty_list_of_preds CLOSE_PAR  { () }
  | NOTES_ATTRIBUTE STRING                { () }
  | other_annotation                      { () }
;



status:
    SAT        { Sat}
  | UNSAT      { Unsat }
  | UNKNOWN    { Unknown }
;




nonempty_list_of_sorts:
    IDENT                         { () }
  | nonempty_list_of_sorts IDENT  { () }
;

list_of_sorts: 
    nonempty_list_of_sorts        { () }
  |                               { () }
;



nonempty_list_of_funs:
    function_declaration                         { () }
  | nonempty_list_of_funs function_declaration   { () }
;

function_declaration:
    OPEN_PAR IDENT nonempty_list_of_sorts list_of_annotations CLOSE_PAR { () }
;




nonempty_list_of_preds:
    predicate_declaration                         { () }
  | nonempty_list_of_preds predicate_declaration  { () }
;

predicate_declaration:
    OPEN_PAR IDENT list_of_sorts list_of_annotations CLOSE_PAR { () }
;





formula:
    atom                                                   { $1 }

  | OPEN_PAR atom nonempty_list_of_annotations CLOSE_PAR   { $2 }


  | OPEN_PAR EQUAL list_of_at_least_two_terms list_of_annotations CLOSE_PAR 
      { mk_list_eq $3 }

  | OPEN_PAR DISTINCT list_of_at_least_two_terms list_of_annotations CLOSE_PAR 
      { mk_distinct $3 }

  | OPEN_PAR LT term term list_of_annotations CLOSE_PAR
      { mk_lt $3 $4 }

  | OPEN_PAR GT term term list_of_annotations CLOSE_PAR
      { mk_gt $3 $4 }

  | OPEN_PAR LEQ term term list_of_annotations CLOSE_PAR
      { mk_leq $3 $4 }

  | OPEN_PAR GEQ term term list_of_annotations CLOSE_PAR
      { mk_geq $3 $4 } 

  | OPEN_PAR NOT formula list_of_annotations CLOSE_PAR
      { mk_not $3 }

  | OPEN_PAR IMPLIES formula formula list_of_annotations CLOSE_PAR 
      { mk_implies $3 $4 }

  | OPEN_PAR IF_THEN_ELSE formula formula formula list_of_annotations CLOSE_PAR
      { mk_ite $3 $4 $5 }

  | OPEN_PAR OR nonempty_list_of_formulas list_of_annotations CLOSE_PAR
      { mk_or_list $3 }

  | OPEN_PAR AND nonempty_list_of_formulas list_of_annotations CLOSE_PAR 
      { mk_and_list $3 }

  | OPEN_PAR XOR nonempty_list_of_formulas list_of_annotations CLOSE_PAR 
      { mk_xor_list $3 }

  | OPEN_PAR IFF nonempty_list_of_formulas list_of_annotations CLOSE_PAR 
      { mk_iff_list $3 }

  | OPEN_PAR LET var_definition formula list_of_annotations CLOSE_PAR
      { remove_term_alias $3; $4 }

  | OPEN_PAR FLET fvar_definition formula list_of_annotations CLOSE_PAR
      { remove_formula_alias $3; $4 }
;

var_definition: 
  OPEN_PAR VAR term CLOSE_PAR      { add_term_alias $2 $3; $2 }
;

fvar_definition:
  OPEN_PAR FVAR formula CLOSE_PAR  { add_formula_alias $2 $3; $2 }
;

nonempty_list_of_formulas: 
    nonempty_list_of_formulas formula   { append_formula $1 $2 }
  | formula                             { unit_formula_list $1 }
;


atom:
    TT      { mk_true }
  | FF      { mk_false }
  | FVAR    { mk_propvar $1 }
  | IDENT   { mk_propvar $1 }
;


term:
    base_term { $1 }

  | OPEN_PAR base_term nonempty_list_of_annotations CLOSE_PAR { $2 }

  | OPEN_PAR PLUS nonempty_list_of_terms list_of_annotations CLOSE_PAR 
      { mk_sum_list $3 }

  | OPEN_PAR MINUS nonempty_list_of_terms list_of_annotations CLOSE_PAR 
      { mk_diff_list $3 }

  | OPEN_PAR TIMES nonempty_list_of_terms list_of_annotations CLOSE_PAR 
      { mk_prod_list $3 }

  | OPEN_PAR DIVIDE term term list_of_annotations CLOSE_PAR 
      { mk_div $3 $4 }

  | OPEN_PAR OPPOSITE term list_of_annotations CLOSE_PAR 
      { mk_opposite $3 }

  | OPEN_PAR ITE formula term term list_of_annotations CLOSE_PAR 
      { mk_cond $3 $4 $5 }
;

nonempty_list_of_terms:
    nonempty_list_of_terms term  { append_term $1 $2 }
  | term                         { unit_term_list $1 }
;

list_of_at_least_two_terms:
    nonempty_list_of_terms term { append_term $1 $2 }
;



base_term:
    VAR     { mk_var $1 }
  | IDENT   { mk_var $1 }
  | NUMBER  { mk_const $1 }
;

%%
