open Formula

(* TODO incorrect wrt variable capture *)

let rec without_quantifiers : source_quantified_formula ->
  source_unquantified_formula = function
  | `Exists(_, f) | `Forall(_, f) -> without_quantifiers f
  | f -> quantified_to_unquantified f;;

let rec extract_quantifiers_blocks_exists_phase :
    source_quantified_formula -> string list list = function
  | `Exists(quantifiers, f) ->
      quantifiers::(extract_quantifiers_blocks_forall_phase f)
  | `Forall _ -> failwith "bad phase"
  | _ -> []

and extract_quantifiers_blocks_forall_phase :
    source_quantified_formula -> string list list = function
  | `Forall(quantifiers, f) ->
      quantifiers::(extract_quantifiers_blocks_exists_phase f)
  | `Exists _ -> failwith "bad phase"
  | _ -> [];;

let simple_generalize_model (context : Formula_yices.linear_translate_context)
    (model : Formula_yices.smt_model) =
  let atoms = ref [] in
    Hashtbl.iter
       (fun (linear_atomic : linear_atomic) yices_expr ->
          match linear_atomic with
            | `Propositional _ -> failwith "propositional"
            | `Greater_eq expr ->
                let polarity =
                  if ((Q.sgn (eval_linear_expression
                                model.Formula_yices.real_model expr)) >= 0)
                  then `GREATER_EQ
                  else `LESS in
                  atoms := (expr, polarity) :: !atoms
       )
       context.Formula_yices.linear_translate_atomic_map; !atoms;;

let do_not_generalize_model (context : Formula_yices.linear_translate_context)
    (model : Formula_yices.smt_model) =
  Array.to_list (Array.mapi
		   (fun i varname ->
		      let v = model.Formula_yices.real_model i in
		 [-1, Z.neg (Q.get_num v); i, Q.get_den v], `EQUAL
	      ) context.Formula_yices.linear_translate_real_map);;

type block_descr = {
  block_width : int;
  block_context : Formula_yices.linear_translate_context;
  (* mutable block_models_found : linear_atomic conjunction list *)
}

let check_sat (lc : linearization_context) (formula : linear_unquantified_formula) : bool =
  let context = Formula_yices.mk_linear_translate_context (Yices.mk_context ()) lc in
  let translated = Formula_yices.translate_linear_formula context formula in
    Yices.assert_expr context.Formula_yices.linear_translate_context translated;
    match Yices.check context.Formula_yices.linear_translate_context with
      | Yices.L_TRUE -> true
      | Yices.L_FALSE -> false
      | Yices.L_UNKNOWN -> failwith "unknown in check_sat";;
    

let rec find_model
    (lc : linearization_context)
    (output_dimension : int)
    (context0 : Formula_yices.dual_linear_translate_context)
    (block_widths_and_contexts : block_descr list)
    (c : (Formula.linear_expression * convex_comparison_op) list) :
    (Formula.linear_expression * convex_comparison_op) list Formula_yices.smt_answer =
  let cf =
    ((Formula_yices.conjuncts_to_conjunction
		 (c :> (Formula.linear_expression * Formula.convex_comparison_op) list))
	      : Formula.linear_atomic Formula.conjunction
	      :> linear_unquantified_formula) in
  match block_widths_and_contexts with
    | [] (* leaf unquantified formula *) ->
	let translated = Formula_yices.translate_linear_formula
	     context0.Formula_yices.positive_context cf in
	Yices.push context0.Formula_yices.positive_context.Formula_yices.linear_translate_context;
	Yices.assert_expr context0.Formula_yices.positive_context.Formula_yices.linear_translate_context translated;
	let model = Formula_yices.get_a_generalized_model lc context0 in
	  Yices.pop context0.Formula_yices.positive_context.Formula_yices.linear_translate_context;
	  (model :> (Formula.linear_expression * convex_comparison_op) list Formula_yices.smt_answer)

    | descr::rest ->
	let translated = Formula_yices.translate_linear_formula descr.block_context cf in
	Yices.push descr.block_context.Formula_yices.linear_translate_context;
	Yices.assert_expr descr.block_context.Formula_yices.linear_translate_context
          translated;
	let perhaps_numeric_model = Formula_yices.linear_solve descr.block_context in
	  Yices.pop descr.block_context.Formula_yices.linear_translate_context;

	  (match perhaps_numeric_model with
	     | Formula_yices.UNSATISFIABLE -> Formula_yices.UNSATISFIABLE
	     | Formula_yices.SATISFIABLE c2 ->
		 let model = simple_generalize_model descr.block_context c2 in
		 (match find_model lc (output_dimension + descr.block_width)
		    context0 rest model with
		      | Formula_yices.UNSATISFIABLE ->
			  let model2 = Minimal.generalize
			    (fun tested_model ->
			       not (is_satisfiable lc (output_dimension + descr.block_width)
				      context0 rest tested_model)) [] model in
			  (* descr.block_models_found <- (Formula_yices.conjuncts_to_conjunction model2)::descr.block_models_found; *)
			  Formula_yices.SATISFIABLE model2

		      | Formula_yices.SATISFIABLE model2 ->
 			  let to_be_removed =
			    Formula_yices.conjuncts_to_formula
			      (Formula_yices.conjuncts_inject
				 (Projection.project_minimize_constraints
 				    model2 output_dimension)) in
			  (* print_string "to be removed: ";
			  Formula_math.print_source_unquantified_formula
			    stdout
			    (unlinearize_unquantified_formula lc
			       to_be_removed);
			    print_newline(); *)
			  Yices.assert_expr
			    descr.block_context.Formula_yices.linear_translate_context
			    (Formula_yices.translate_linear_formula descr.block_context
			    (`Not to_be_removed));
			  find_model lc output_dimension context0
			    block_widths_and_contexts c))

and is_satisfiable
    (lc : linearization_context)
    (output_dimension : int)
    (context0 : Formula_yices.dual_linear_translate_context)
    (block_widths_and_contexts : block_descr list)
    (c : (Formula.linear_expression * convex_comparison_op) list) : bool=
  let cf =
    ((Formula_yices.conjuncts_to_conjunction
		 (c :> (Formula.linear_expression * Formula.convex_comparison_op) list))
	      : Formula.linear_atomic Formula.conjunction
	      :> linear_unquantified_formula) in
  match block_widths_and_contexts with
    | [] (* leaf unquantified formula *) ->
	let translated = Formula_yices.translate_linear_formula
	     context0.Formula_yices.positive_context cf in
	Yices.push context0.Formula_yices.positive_context.Formula_yices.linear_translate_context;
	Yices.assert_expr context0.Formula_yices.positive_context.Formula_yices.linear_translate_context translated;
	let answer = Yices.check context0.Formula_yices.positive_context.Formula_yices.linear_translate_context in
	  Yices.pop context0.Formula_yices.positive_context.Formula_yices.linear_translate_context;
	  (match answer with
	     | Yices.L_TRUE->true
	     | Yices.L_FALSE -> false
	     | Yices.L_UNKNOWN -> failwith "undecidable model???")

    | descr::rest ->
	(* if check_sat lc (`And [cf; `Or (descr.block_models_found : linear_atomic conjunction list :> linear_unquantified_formula list)])
        then true
        else *)

	let translated = Formula_yices.translate_linear_formula descr.block_context cf in
	Yices.push descr.block_context.Formula_yices.linear_translate_context;
	Yices.assert_expr descr.block_context.Formula_yices.linear_translate_context
          translated;
	let perhaps_numeric_model = Formula_yices.linear_solve descr.block_context in
	  Yices.pop descr.block_context.Formula_yices.linear_translate_context;

	  (match perhaps_numeric_model with
	     | Formula_yices.UNSATISFIABLE -> false
	     | Formula_yices.SATISFIABLE c2 ->
		 (*if not (is_satisfiable lc (output_dimension + descr.block_width) context0 rest
		   (do_not_generalize_model descr.block_context c2))
		 then true
		 else*)
		 let model = simple_generalize_model descr.block_context c2 in
		 (match find_model lc  (output_dimension + descr.block_width)
		    context0 rest model with
		      | Formula_yices.UNSATISFIABLE -> true

		      | Formula_yices.SATISFIABLE model2 ->
 			  let to_be_removed =
			    Formula_yices.conjuncts_to_formula
			      (Formula_yices.conjuncts_inject
				 (Projection.project_minimize_constraints
 				    model2 output_dimension)) in

			  Yices.assert_expr
			    descr.block_context.Formula_yices.linear_translate_context
			    (Formula_yices.translate_linear_formula descr.block_context
			    (`Not to_be_removed));
			  is_satisfiable lc output_dimension context0
			    block_widths_and_contexts c));;

let negate_appropriately n f =
  if (n mod 2)==1
  then mk_not f
  else f;;

let linearization_context_from_array (a : string array) =
  let hashtable = Hashtbl.create (Array.length a) in
    Array.iteri (fun i x -> Hashtbl.add hashtable x i) a;
    {
      trans_var_table = hashtable;
      trans_nr_vars = Array.length a;
      trans_var_rev_table = a
    };;

let add_blocking_to_root
    (context0 : Formula_yices.dual_linear_translate_context)
    (block_widths_and_contexts : block_descr list)
    (blocking_clause : linear_unquantified_formula) =
  match block_widths_and_contexts with
    | [] -> Yices.assert_expr
	context0.Formula_yices.positive_context.Formula_yices.linear_translate_context
	  (Formula_yices.translate_linear_formula
	     context0.Formula_yices.positive_context blocking_clause)
    | descr::_ ->
	Yices.assert_expr descr.block_context.Formula_yices.linear_translate_context
	  (Formula_yices.translate_linear_formula descr.block_context blocking_clause);;
    
let rec elim_to_dnf (lc : linearization_context)
    (output_dimension : int)
    (context0 : Formula_yices.dual_linear_translate_context)
    (block_widths_and_contexts : block_descr list)
    (already : (Formula.linear_expression * convex_comparison_op) list list) =
  match find_model lc output_dimension context0 block_widths_and_contexts [] with
    | Formula_yices.UNSATISFIABLE -> already
    | Formula_yices.SATISFIABLE model ->
 	let to_be_removed =
	  Formula_yices.conjuncts_to_formula model in
	  add_blocking_to_root context0 block_widths_and_contexts
	    (`Not  to_be_removed);
	  elim_to_dnf lc output_dimension context0 block_widths_and_contexts
	    (model::already);;

let eliminate_sub (modulo_theory : source_unquantified_formula)
    (f : source_quantified_formula) :
    source_unquantified_formula =
  let unq = without_quantifiers f
  and free_variables = StringSet.elements
    (StringSet.union
       (collect_free_variables_in_formula
	  StringSet.empty StringSet.empty f)
       (collect_free_variables_in_formula
	  StringSet.empty StringSet.empty (Formula.unquantified_to_quantified modulo_theory))) in
  let output_dimension = List.length free_variables in
  let quantifier_blocks = extract_quantifiers_blocks_forall_phase f in
    
  let lc = linearization_context_from_array
    (Array.of_list (List.concat (free_variables::quantifier_blocks))) in

  let modulo_theory_l = linearize_unquantified_formula lc modulo_theory in
  let context0 = Formula_yices.mk_dual_context lc
    modulo_theory_l
    (linearize_unquantified_formula lc 
       (negate_appropriately (List.length quantifier_blocks) unq)) in

  let block_widths_and_contexts =
    List.map
      (fun block ->
	 let context = Formula_yices.mk_linear_translate_context
               (Yices.mk_context ()) lc in
	   Yices.assert_expr context.Formula_yices.linear_translate_context
	     (Formula_yices.translate_linear_formula context modulo_theory_l);
	 { block_width = List.length block;
	   (* block_models_found = []; *)
	   block_context = context}) quantifier_blocks  in
  let dnf = elim_to_dnf lc output_dimension context0 block_widths_and_contexts [] in
    unlinearize_unquantified_formula lc
      (`Or
	 ((List.map Formula_yices.conjuncts_to_conjunction dnf)
	 : Formula.linear_atomic Formula.conjunction list
	 :> linear_unquantified_formula list));;

let eliminate modulo_theory (f : source_quantified_formula) :
    source_unquantified_formula =
   match f with
     | `Forall(_, _) -> eliminate_sub modulo_theory f
     | _ -> eliminate_sub modulo_theory (`Forall([],f));;
