module StringSet = Set.Make(String);;

let formula_true = (`And [])
and formula_false = (`Or [])

type 'atomic unquantified_formula =
  [ `And of 'atomic unquantified_formula list
  | `Or of 'atomic unquantified_formula list
  | `Xor of 'atomic unquantified_formula list
  | `Not of 'atomic unquantified_formula
  | `Ite of 'atomic unquantified_formula * 'atomic unquantified_formula * 'atomic unquantified_formula
  | `Atomic of 'atomic ];;

type 'atomic nnf_unquantified_formula =
  [ `And of 'atomic nnf_unquantified_formula list
  | `Or of 'atomic nnf_unquantified_formula list
  | `Atomic of 'atomic
  | `Not of [`Atomic of 'atomic] ];;

let expand_ite c f1 f2 = (`Or[`And [c; f1]; `And[`Not c; f2]])
let rec to_nnf_unquantified_formula : 'atomic unquantified_formula -> 'atomic nnf_unquantified_formula = function
  | (`Atomic _) | (`Not (`Atomic _)) as w -> w
  | `Not x -> to_nnf_unquantified_formula_neg x
  | `And l  -> `And(List.map to_nnf_unquantified_formula l)
  | `Or l -> `Or(List.map to_nnf_unquantified_formula l)
  | `Ite(c, f1, f2) -> to_nnf_unquantified_formula (expand_ite c f1 f2)
  | `Xor _ -> failwith "xor in nnf"
(*
  | `Xor[] -> `Or[]
  | `Xor l -> `Xor (List.map to_nnf_unquantified_formula l)
*)

and to_nnf_unquantified_formula_neg : 'atomic unquantified_formula -> 'atomic nnf_unquantified_formula = function
  | (`Atomic _) as w -> `Not w
  | `Not ((`Atomic _) as w) -> w
  | `Not x -> to_nnf_unquantified_formula x
  | `And l -> `Or(List.map to_nnf_unquantified_formula_neg l)
  | `Or l -> `And(List.map to_nnf_unquantified_formula_neg l)
  | `Ite(c, f1, f2) -> to_nnf_unquantified_formula_neg (expand_ite c f1 f2)
  | `Xor _ -> failwith "xor in nnf"
(*
  | `Xor[] -> `And[]
  | `Xor(h::t) -> `Xor((to_nnf_unquantified_formula_neg h)::
			 (List.map to_nnf_unquantified_formula t))
*);;

let mk_not = function
  | `Not (`And []) -> `Or []
  | `Not (`Or []) -> `And []
  | `Not x -> x
  | x -> `Not x;;

let rec collapse_duplicate = function
  | ([] | [_]) as w -> w
  | h1::(h2::_ as t) ->
      if h1=h2
      then collapse_duplicate t
      else h1::(collapse_duplicate t);;
 
let simplify_vec l = collapse_duplicate (List.sort Pervasives.compare l)

let mk_v_or1 = function
  | [h] -> h
  | x -> `Or x;;

let mk_v_or =
  let rec mk_v_or_rec already = function
    | [] -> mk_v_or1 already
    | (`Or l)::t -> mk_v_or_rec already (List.append l t)
    | ((`And []) as absorber)::_ -> absorber
    | h::t -> mk_v_or_rec (h::already) t
  in fun x -> mk_v_or_rec [] (simplify_vec x);;

let mk_v_and1 = function
  | [h] -> h
  | x -> `And x;;

let mk_v_and =
  let rec mk_v_and_rec already = function
    | [] -> mk_v_and1 already
    | (`And l)::t -> mk_v_and_rec already (List.append l t)
    | ((`Or []) as absorber)::_ -> absorber
    | h::t -> mk_v_and_rec (h::already) t
  in fun x -> mk_v_and_rec [] (simplify_vec x);;

let mk_v_xor =
  let rec mk_v_xor_rec already = function
    | [] -> `Xor already
    | (`Xor l)::t -> mk_v_xor_rec already (List.append l t)
    | h::t -> mk_v_xor_rec (h::already) t
  in fun x -> mk_v_xor_rec [] x;;

let mk_or x y = mk_v_or [x; y];;
let mk_xor x y = mk_v_xor [x; y];;
let mk_and x y = mk_v_and [x; y];;

let mk_implies x y = mk_or (mk_not x) y

let map_atomic_nnf_unquantifier_formula f g =
  let rec mapper : 'atomic nnf_unquantified_formula
    ->'atomic nnf_unquantified_formula  = function
  | `And l -> mk_v_and (List.map mapper l)
  | `Or l -> mk_v_or (List.map mapper l)
  | `Atomic x -> f x
  | `Not(`Atomic x) -> g x in
  fun x -> mapper x;;

type 'atomic literal = [`Atomic of 'atomic | `Not of [`Atomic of 'atomic]];;
type 'atomic conjunction = [`And of 'atomic literal list];;
type 'atomic dnf_formula = [ `Or of 'atomic conjunction list];; 

type ('variable, 'atomic) quantified_formula =
  [ `And of ('variable, 'atomic) quantified_formula list
  | `Or of ('variable, 'atomic) quantified_formula list
  | `Xor of ('variable, 'atomic) quantified_formula list
  | `Not of ('variable, 'atomic) quantified_formula
  | `Ite of ('variable, 'atomic) quantified_formula * ('variable, 'atomic) quantified_formula * ('variable, 'atomic) quantified_formula 
  | `Atomic of 'atomic
  | `Forall of 'variable list * ('variable, 'atomic) quantified_formula
  | `Exists of 'variable list * ('variable, 'atomic) quantified_formula ];;

let add_list_to_string_set = List.fold_right StringSet.add

type ('variable, 'atomic) nnf_quantified_formula =
  [ `And of ('variable, 'atomic) nnf_quantified_formula list
  | `Or of ('variable, 'atomic) nnf_quantified_formula list
  | `Xor of ('variable, 'atomic) nnf_quantified_formula list
  | `Atomic of 'atomic
  | `Not of [`Atomic of 'atomic]
  | `Forall of 'variable list * ('variable, 'atomic) nnf_quantified_formula
  | `Exists of 'variable list * ('variable, 'atomic) nnf_quantified_formula ];;

let rec to_nnf_quantified_formula : ('variable, 'atomic) quantified_formula ->
  ('variable, 'atomic) nnf_quantified_formula = function
  | (`Atomic _) | (`Not (`Atomic _)) as w -> w
  | `Not x -> to_nnf_quantified_formula_neg x
  | `And l  -> `And(List.map to_nnf_quantified_formula l)
  | `Or l -> `Or(List.map to_nnf_quantified_formula l)
  | `Xor[] -> `Or[]
  | `Xor l -> `Xor(List.map to_nnf_quantified_formula l)
  | `Ite(c, f1, f2) -> to_nnf_quantified_formula (expand_ite c f1 f2)
  | `Exists(v, f) -> `Exists(v, to_nnf_quantified_formula f)
  | `Forall(v, f) -> `Forall(v, to_nnf_quantified_formula f)

and to_nnf_quantified_formula_neg : ('variable, 'atomic) quantified_formula ->
  ('variable, 'atomic) nnf_quantified_formula = function
  | (`Atomic _) as w -> `Not w
  | `Not ((`Atomic _) as w) -> w
  | `Not x -> to_nnf_quantified_formula x
  | `And l -> `Or(List.map to_nnf_quantified_formula_neg l)
  | `Or l -> `And(List.map to_nnf_quantified_formula_neg l)
  | `Xor[] -> `And[]
  | `Xor(h::t) -> `Xor((to_nnf_quantified_formula_neg h)::
			 (List.map to_nnf_quantified_formula t))
  | `Ite(c, f1, f2) -> to_nnf_quantified_formula_neg (expand_ite c f1 f2)
  | `Exists(v, f) -> `Forall(v, to_nnf_quantified_formula_neg f)
  | `Forall(v, f) -> `Exists(v, to_nnf_quantified_formula_neg f);;

let mk_exists var = function
  | `Exists(vars, f) -> `Exists((var::vars), f)
  | f -> `Exists([var], f);;

let mk_forall var = function
  | `Forall(vars, f) -> `Forall((var::vars), f)
  | f -> `Forall([var], f);;

let unquantified_to_quantified f = (f : 'atomic unquantified_formula :> ('variable, 'atomic) quantified_formula);;

exception Quantified;;

let rec quantified_to_unquantified (f : ('variable, 'atomic) quantified_formula) : 'atomic unquantified_formula =
  match f with
    | `Forall _ | `Exists _ -> raise Quantified
    | `And l -> `And (List.map quantified_to_unquantified l)
    | `Or l -> `Or (List.map quantified_to_unquantified l)
    | `Xor l -> `Xor (List.map quantified_to_unquantified l)
    | `Ite(c, f1, f2) -> `Ite((quantified_to_unquantified c),
			      (quantified_to_unquantified f1),
			      (quantified_to_unquantified f2))
    | `Not l -> `Not (quantified_to_unquantified l)
    | (`Atomic _) as w -> w;;

let rec quantified_formula_size :
    ('variable, 'atomic) quantified_formula -> int = function
      | `Atomic _ -> 1
      | `And l | `Or l | `Xor l ->
	  List.fold_right (fun f previous ->
			     previous + (quantified_formula_size f)) l 1
      | `Ite(c, f1, f2) ->
	  (quantified_formula_size c) +
	  (quantified_formula_size f1) +
          (quantified_formula_size f2)
      | `Not f | `Forall(_, f) | `Exists(_, f) ->
	  1 + (quantified_formula_size f);;

let unquantified_formula_size (f : 'atomic unquantified_formula) =
  quantified_formula_size (unquantified_to_quantified f);;

let print_formula (print_atomic : out_channel -> 'atomic -> unit)
    (print_variable : out_channel -> 'variable -> unit)
    (channel : out_channel) :
    ('variable, 'atomic) quantified_formula -> unit =
  let outs = output_string channel in
  let outvl = function
    | [] -> outs "{}"
    | [v] -> print_variable channel v
    | h::t -> print_variable channel h;
        List.iter (fun v -> outs ", "; print_variable channel v) t in
  let rec prf : ('variable, 'atomic) quantified_formula -> unit = function
    | `And [] -> outs "True"
    | `And l -> outs "And[";
      print_list l;
      outs "]"
    | `Or [] -> outs "False"
    | `Or l -> outs "Or[";
	print_list l;
	outs "]"
    | `Xor l -> outs "Xor[";
	print_list l;
	outs "]"
    | `Not w -> outs "Not[";
	prf w;
	outs "]"
    | `Ite(c, f1, f2) -> outs "If[";
        prf c;
        outs ", ";
        prf f1;
        outs ", ";
        prf f2;
        outs "]"
    | `Atomic a -> print_atomic channel a
    | `Exists([v], f) -> outs "Exists[";
	outs v;
	outs ", ";
	prf f;
	outs "]"
    | `Forall([v], f) -> outs "ForAll[";
	outs v;
	outs ", ";
	prf f;
	outs "]"
    | `Exists(vl, f) -> outs "Exists[{";
	outvl vl;
	outs "}, ";
	prf f;
	outs "]"
    | `Forall(vl, f) -> outs "ForAll[{";
	outvl vl;
	outs "}, ";
	prf f;
	outs "]"
  and print_list : ('variable, 'atomic) quantified_formula list -> unit= function
  | [] -> ()
  | [h] -> prf h
  | h::t -> prf h; outs ", "; print_list t

  in prf;;

let rec eval_formula (eval_atomic : 'atomic -> bool) :
  'atomic unquantified_formula -> bool = function
    | `And l -> List.fold_left (fun previous f ->
	let x = eval_formula eval_atomic f in previous && x) true l
    | `Or l -> List.fold_left (fun previous f ->
	let x = eval_formula eval_atomic f in previous || x) false l
    | `Xor l -> List.fold_left (fun previous f ->
	let x = eval_formula eval_atomic f in previous <> x) false l
    | `Not f -> not (eval_formula eval_atomic f)
    | `Ite(c, f1, f2) ->
	if (eval_formula eval_atomic c)
	then (eval_formula eval_atomic f1)
	else (eval_formula eval_atomic f2)
    | `Atomic a -> eval_atomic a;;
     
type source_expression =
    [ `Sum of source_expression list
    | `Product of source_expression list
    | `Neg of source_expression
    | `Quotient of source_expression * source_expression
    | `Integer of Z.t
    | `Variable of string ];;

let zero_expression = `Sum [];;

let mk_sum x y =
  match x, y with
    | (`Sum lx), (`Sum ly) -> `Sum(List.append lx ly)
    | (`Sum l), a | a, (`Sum l) -> `Sum(a::l)
    | ax, ay -> `Sum [ax; ay];;

let mk_product x y =
  match x, y with
    | (`Product lx), (`Product ly) -> `Product(List.append lx ly)
    | (`Product l), a | a, (`Product l) -> `Product(a::l)
    | ax, ay -> `Product [ax; ay];;

let mk_neg = function
  | `Neg x -> x
  | x -> `Neg x;;

let rec eval_source_expression (eval_variable : string -> Q.t) :
  source_expression -> Q.t = function
    | `Sum l -> List.fold_left (fun previous f ->
       Q.add previous (eval_source_expression eval_variable f)) Q.zero l
    | `Product l -> List.fold_left (fun previous f ->
       Q.mul previous (eval_source_expression eval_variable f)) Q.one l
    | `Neg f -> Q.neg (eval_source_expression eval_variable f)
    | `Quotient(f1, f2) -> Q.div (eval_source_expression eval_variable f1)
         (eval_source_expression eval_variable f2)
    | `Integer x -> Q.from_z x
    | `Variable x -> eval_variable x;;

type convex_comparison_op = 
    [ `LESS | `GREATER | `LESS_EQ | `GREATER_EQ | `EQUAL ];;

type comparison_op =
    [ `LESS | `GREATER | `LESS_EQ | `GREATER_EQ | `EQUAL | `UNEQUAL ];;

type source_atomic =
    [ `Propositional of string
    | `Comparison of source_expression * comparison_op * source_expression ];;

let eval_source_atomic (eval_real_variable : string -> Q.t)
  (eval_prop_variable : string -> bool) : source_atomic -> bool = function
    | `Propositional x -> eval_prop_variable x
    | `Comparison(lhs, op, rhs) ->
	let cmp = Q.compare (eval_source_expression eval_real_variable lhs)
	  (eval_source_expression eval_real_variable rhs) in
	  (match op with
	     | `LESS -> cmp < 0
	     | `LESS_EQ -> cmp <= 0
	     | `GREATER -> cmp > 0
	     | `GREATER_EQ -> cmp >= 0
	     | `EQUAL -> cmp = 0
	     | `UNEQUAL -> cmp <> 0);;
	
type source_quantified_formula = (string, source_atomic) quantified_formula;;
type source_unquantified_formula = source_atomic unquantified_formula;;
type source_dnf_formula = source_atomic dnf_formula;;

let eval_source_formula (eval_real_variable : string -> Q.t)
  (eval_prop_variable : string -> bool) : source_unquantified_formula -> bool =
  eval_formula (eval_source_atomic eval_real_variable eval_prop_variable);;

type linear_expression = (int * Z.t) list;;

let linear_expression_get_coefficient l x =
  let r = List.assoc x l in
    assert((Z.sgn r) != 0);
    r;;

type q_linear_expression = {
  expr : linear_expression;
  num : Z.t;
  den : Z.t };;

let zero_q_linear_expression = {
  expr = [];
  num = Z.one;
  den = Z.one };;

let linear_expr_ok =
  let rec check_rec = function
    | [] -> true
    | [v, x] -> v >= 0 && (Z.sgn x) <> 0
    | (v1, x1) :: (((v2, _) :: _) as t) ->
	(v1 >= 0) && (Z.sgn x1) <> 0 && (v1 < v2) && (check_rec t)
  in function
    | (-1, _) :: t -> check_rec t
    | t -> check_rec t;;

let multiply_add_linear_expressions (c1 : Z.t) (c2 : Z.t) =
  let rec mul_add_rev (rest : linear_expression)
      (l1 : linear_expression) (l2 : linear_expression) : linear_expression =
  match l1, l2 with
    | [], [] -> rest
    | ((v1, w1)::t1), [] ->
	let coeff = Z.mul c1 w1 in
	  mul_add_rev
	    (if (Z.sgn coeff = 0) then rest else (v1, coeff)::rest) t1 []

    | [], ((v2, w2)::t2) ->
	let coeff = Z.mul c2 w2 in
	  mul_add_rev
	    (if (Z.sgn coeff = 0) then rest else (v2, coeff)::rest) [] t2

    | ((v1, w1)::t1), ((v2, w2)::t2) ->
	if v1 = v2
        then let coeff = Z.add (Z.mul c1 w1) (Z.mul c2 w2) in
	  mul_add_rev
	    (if (Z.sgn coeff = 0) then rest else (v1, coeff)::rest) t1 t2
	else
	  if v1 < v2
	  then
	    let coeff = Z.mul c1 w1 in
	      mul_add_rev
		(if (Z.sgn coeff = 0) then rest else (v1, coeff)::rest) t1 l2
	  else
	    let coeff = Z.mul c2 w2 in
	      mul_add_rev
		(if (Z.sgn coeff = 0) then rest else (v2, coeff)::rest) l1 t2

  in fun l1 l2 ->
    let l = List.rev (mul_add_rev [] l1 l2) in
    assert(linear_expr_ok l);
    l;;
	  
let linear_expression_gcd (l : linear_expression) =
  try
    match l with
      | [] -> Z.one
      | (first_var_id, first_coeff)::rest ->
	  List.fold_left
	    (fun before (var_id, coeff) ->
	       if (Z.cmp before Z.one) = 0
	       then raise Exit
	       else Z.abs (Z.gcd before coeff))
	    (Z.abs first_coeff) rest
  with Exit -> Z.one;;

let linear_expression_divexact (expr : linear_expression) (divisor : Z.t) =
  List.map (fun (var_id, coeff) -> (var_id, (Z.divexact coeff divisor))) expr;;

let gen_q_linear_expression_positive expr den =
  let num = linear_expression_gcd expr in
  let num_den_gcd = Z.gcd num den in
  let reduced_num = Z.divexact num num_den_gcd
  and reduced_den = Z.divexact den num_den_gcd
  and reduced_expr = linear_expression_divexact expr num in
  let ret =
    { expr = reduced_expr;
      num = reduced_num;
      den = reduced_den } in
  (* print_endline "gen";
  print_q_linear_expression stdout (fun id -> Printf.sprintf "v%d" id) ret;
  print_newline(); *)
  ret ;;

let linear_expression_neg (l : linear_expression) =
  List.map (fun (var_id, coeff) -> (var_id, Z.neg coeff)) l;;

let gen_q_linear_expression expr den =
  let sign = Z.sgn den in
  if sign < 0
  then gen_q_linear_expression_positive (linear_expression_neg expr) (Z.neg den)
  else if sign > 0
  then gen_q_linear_expression_positive expr den
  else assert false;;

let q_linear_expression_add
  (l1 : q_linear_expression) (l2 : q_linear_expression) =
  (*print_endline "add";
  print_q_linear_expression stdout (fun id -> Printf.sprintf "v%d" id) l1;
  print_newline();
  print_q_linear_expression stdout (fun id -> Printf.sprintf "v%d" id) l2;
  print_newline();*)
  let den = Z.lcm l1.den l2.den in
  let s1 = Z.mul (Z.divexact den l1.den) l1.num
  and s2 = Z.mul (Z.divexact den l2.den) l2.num in
  let sum = multiply_add_linear_expressions s1 s2 l1.expr l2.expr in
    gen_q_linear_expression sum den;;

let q_linear_expression_sub
  (l1 : q_linear_expression) (l2 : q_linear_expression) =
  let den = Z.lcm l1.den l2.den in
  let s1 = Z.mul (Z.divexact den l1.den) l1.num
  and s2 = Z.neg (Z.mul (Z.divexact den l2.den) l2.num) in
  let sum = multiply_add_linear_expressions s1 s2 l1.expr l2.expr in
    gen_q_linear_expression sum den;;

exception Nonlinear
exception Not_a_constant

let linear_expression_scale (c : Z.t) (e : linear_expression) =
  if (Z.sgn c) = 0
  then []
  else List.map (fun (var_id, w) -> (var_id, (Z.mul w c))) e;;

let linear_expression_mul
  (l1 : linear_expression) (l2 : linear_expression) =
  match l1, l2 with
    | [], _ | _, [] -> []
    | [-1, c], e | e, [-1, c] -> linear_expression_scale c e
    | _ -> raise Nonlinear

let q_linear_expression_reduce expr num den =
  let common = Z.gcd num den in
  { expr = expr;
    num = Z.divexact num common ;
    den = Z.divexact den common };;

let q_linear_expression_mul
  (l1 : q_linear_expression) (l2 : q_linear_expression) =
  q_linear_expression_reduce (linear_expression_mul l1.expr l2.expr)
    (Z.mul l1.num l2.num)
    (Z.mul l1.den l2.den);;

type sign = ZERO | NEGATIVE | POSITIVE

let minus_one = Z.neg Z.one;;

let extract_constant_sign (l : linear_expression) =
  match l with
    | [] -> ZERO
    | [-1, w] when (Z.compare w minus_one) = 0 -> NEGATIVE
    | [-1, w] when (Z.compare w Z.one) = 0 -> POSITIVE
    | [-1, _] -> failwith "non-reduced constant"
    | _::_ -> raise Not_a_constant;;

let q_linear_expression_div
    (l1 : q_linear_expression) (l2 : q_linear_expression) =
  match extract_constant_sign l2.expr with
    | ZERO -> raise Division_by_zero
    | NEGATIVE -> 
	q_linear_expression_reduce
	  (linear_expression_neg l1.expr)
	  (Z.mul l2.den l1.num)
	  (Z.mul l2.num l1.den)
    | POSITIVE ->
	q_linear_expression_reduce
	  l1.expr
	  (Z.mul l2.den l1.num)
	  (Z.mul l2.num l1.den);;
  
let q_linear_expression_var (var_id : int) = 
  assert(var_id >= 0);
  { expr = [var_id, Z.one];
    num = Z.one;
    den = Z.one };;

let q_linear_expression_neg x = {x with expr = linear_expression_neg x.expr};;

let q_linear_expression_zero =
  { expr = [];
    num = Z.one;
    den = Z.one };;

let q_linear_expression_const (value : Z.t) = 
  match Z.sgn value with
    | 0 -> { expr = [];
	     num = Z.one;
	     den = Z.one }
    | n when n < 0 ->
	{ expr = [-1, Z.neg Z.one];
	  num = Z.neg value;
	  den = Z.one }
    | _ ->
	{ expr = [-1, Z.one];
	  num = value;
	  den = Z.one };;

let q_linear_expression_one = q_linear_expression_const Z.one;;

let rec filter_out_from_linear_expression (var_id : int) :
    linear_expression -> linear_expression = function
      | [] -> []
      | ((var_id', _)::_) as t when var_id' > var_id -> t
      | (var_id', _)::t when var_id' = var_id -> t
      | h::t -> h::(filter_out_from_linear_expression var_id t);;

let substitute_in_q_linear_expression (var_id : int)
    (replacement : q_linear_expression) (old : q_linear_expression) :
    q_linear_expression =
  try
    let coeff = linear_expression_get_coefficient old.expr var_id in
      q_linear_expression_add
	{ old with expr = filter_out_from_linear_expression var_id old.expr }
	{ replacement with num = Z.mul replacement.num coeff }
  with Not_found -> old;;

type linear_atomic =
    [ `Greater_eq of linear_expression
    | `Propositional of string ];;

type linear_quantified_formula = (string, linear_atomic) quantified_formula;;
type linear_unquantified_formula = linear_atomic unquantified_formula;;

let gen_greater_eq = function
  | [] -> formula_true
  | [-1, x] ->
      if (Z.sgn x) >= 0 then formula_true else formula_false
  | t -> `Atomic (`Greater_eq t);;

let gen_less = function
  | [] -> formula_false
  | [-1, x] ->
      if (Z.sgn x) >= 0 then formula_false else formula_true
  | t -> `Not (`Atomic (`Greater_eq t));;

let substitute_in_linear_atomic (var_id : int)
    (replacement : q_linear_expression) =
  function
    | (`Propositional x) as w -> `Atomic w
    | `Greater_eq e ->
	let e' = substitute_in_q_linear_expression var_id
	  replacement { expr=e; num=Z.one; den=Z.one } in
	  assert((Z.sgn e'.num) > 0);
	  assert((Z.sgn e'.den) > 0);
	  gen_greater_eq e'.expr;;

let substitute_in_not_linear_atomic (var_id : int)
    (replacement : q_linear_expression) =
  function
    | (`Propositional x) as w -> `Not (`Atomic w)
    | `Greater_eq e ->
	let e' = substitute_in_q_linear_expression var_id
	  replacement { expr=e; num=Z.one; den=Z.one } in
	  assert((Z.sgn e'.num) > 0);
	  assert((Z.sgn e'.den) > 0);
	  gen_less e'.expr;;

let substitute_in_nnf_linear_formula (var_id : int)
    (replacement : q_linear_expression) =
  let rec subst_rec :linear_atomic nnf_unquantified_formula ->
    linear_atomic nnf_unquantified_formula = function
      | `And l -> mk_v_and (List.map subst_rec l)
      | `Or l -> mk_v_or (List.map subst_rec l)
      | `Atomic f -> substitute_in_linear_atomic var_id replacement f
      | `Not(`Atomic f) ->
	  substitute_in_not_linear_atomic var_id replacement f in
    fun x -> subst_rec x;;

type linearization_context =
  { mutable trans_nr_vars : int;
    trans_var_table : (string, int) Hashtbl.t;
    mutable trans_var_rev_table : string array };;

let empty_linearization_context () =
  { trans_nr_vars = 0;
    trans_var_table = Hashtbl.create 100;
    trans_var_rev_table = [| |] };;

let linearize_real_var (ct : linearization_context) (v : string) =
  try Hashtbl.find ct.trans_var_table v
  with Not_found ->
    let id = ct.trans_nr_vars in
      Hashtbl.add ct.trans_var_table v id;
      ct.trans_nr_vars <- id + 1;
      id;;

let rec linearize_expression (ct : linearization_context) :
    source_expression -> q_linear_expression = function
      | `Sum l ->
	  List.fold_left (fun previous e ->
	    q_linear_expression_add previous (linearize_expression ct e))
            q_linear_expression_zero l
      | `Product l ->
	  List.fold_left (fun previous e ->
	    q_linear_expression_mul previous (linearize_expression ct e))
            q_linear_expression_one l

      | `Neg e -> q_linear_expression_neg (linearize_expression ct e)
      | `Quotient(e1, e2) -> q_linear_expression_div
	  (linearize_expression ct e1)
	    (linearize_expression ct e2)
      | `Integer x -> q_linear_expression_const x
      | `Variable x -> q_linear_expression_var (linearize_real_var ct x);;

let linearize_atomic (ct : linearization_context) = function
      | (`Propositional _) as w -> `Atomic w
      | `Comparison(lhs, op, rhs) ->
	  let delta = q_linear_expression_sub (linearize_expression ct lhs)
	    (linearize_expression ct rhs) in
	    assert((Z.sgn delta.num) > 0);
	    assert((Z.sgn delta.den) > 0);
	    (match op with
	       | `GREATER_EQ -> `Atomic (`Greater_eq delta.expr)
	       | `LESS -> `Not (`Atomic (`Greater_eq delta.expr))
	       | `LESS_EQ ->
		   `Atomic (`Greater_eq (linear_expression_neg delta.expr))
	       | `GREATER ->
		   `Not(`Atomic (`Greater_eq
                      (linear_expression_neg delta.expr)))
	       | `EQUAL ->
		   `And [`Atomic (`Greater_eq delta.expr);
	             `Atomic (`Greater_eq (linear_expression_neg delta.expr))]
	       | `UNEQUAL -> `Not(
		   `And [`Atomic (`Greater_eq delta.expr);
	             `Atomic (`Greater_eq (linear_expression_neg delta.expr))]
		 ));;

let rec linearize_unquantified_formula (ct : linearization_context) :
    source_unquantified_formula -> linear_unquantified_formula =
  function
    | `And l -> `And (List.map (linearize_unquantified_formula ct) l)
    | `Or l -> `Or (List.map (linearize_unquantified_formula ct) l)
    | `Xor l -> `Xor (List.map (linearize_unquantified_formula ct) l)
    | `Not f -> `Not(linearize_unquantified_formula ct f)
    | `Ite(c, f1, f2) -> `Ite((linearize_unquantified_formula ct c),
			      (linearize_unquantified_formula ct f1),
			      (linearize_unquantified_formula ct f2))
    | `Atomic f -> linearize_atomic ct f;;

let unlinearize_expression (ct : linearization_context)
    (l : linear_expression) : source_expression =
  let summands =
    List.map
	 (fun (var_id, coeff) ->
	    assert(var_id >= -1);
	    if var_id = -1
	    then
	      `Integer coeff
            else
	      let v = `Variable (ct.trans_var_rev_table.(var_id)) in
		if (Z.cmp coeff Z.one) = 0
		then v
		else
		  if (Z.cmp coeff (Z.neg Z.one)) = 0
		  then `Neg v
		  else `Product [`Integer coeff; v]) l in
    match summands with
      | [] -> zero_expression
      | [h] -> h
      | l -> `Sum l;;


let unlinearize_dnf_formula (ct : linearization_context) :
    linear_atomic dnf_formula -> source_atomic dnf_formula =
  function `Or(disjuncts) ->
    `Or(List.map (function `And(conjuncts) ->
      `And(List.map
        (function
           | (`Atomic (`Propositional p))
           | (`Not (`Atomic (`Propositional p))) as w -> w
	   | `Atomic (`Greater_eq f) ->
	     `Atomic (`Comparison(unlinearize_expression ct f, `GREATER_EQ, zero_expression))
	   | `Not (`Atomic (`Greater_eq f)) ->
	`Atomic (`Comparison(unlinearize_expression ct f, `LESS, zero_expression))
	) conjuncts)) disjuncts);;
		   
let rec unlinearize_unquantified_formula (ct : linearization_context) :
    linear_unquantified_formula -> source_unquantified_formula =
  function
    | `And l -> `And (List.map (unlinearize_unquantified_formula ct) l)
    | `Or l -> `Or (List.map (unlinearize_unquantified_formula ct) l)
    | `Xor l -> `Xor (List.map (unlinearize_unquantified_formula ct) l)
    | (`Atomic (`Propositional p))
    | (`Not (`Atomic (`Propositional p))) as w -> w
    | `Atomic (`Greater_eq f) ->
	`Atomic (`Comparison(unlinearize_expression ct f, `GREATER_EQ, zero_expression))
    | `Not (`Atomic (`Greater_eq f)) ->
	`Atomic (`Comparison(unlinearize_expression ct f, `LESS, zero_expression))
    | `Not f -> `Not (unlinearize_unquantified_formula ct f)
    | `Ite(c, f1, f2) -> `Ite((unlinearize_unquantified_formula ct c),
			      (unlinearize_unquantified_formula ct f1),
			      (unlinearize_unquantified_formula ct f2));;

let eval_linear_expression (eval_variable : int -> Q.t)
    (l : linear_expression) : Q.t =
  let r = List.fold_left (fun previous (id, coeff) ->
    Q.add previous 
      (if id >= 0
       then Q.mul (Q.from_z coeff) (eval_variable id)
       else Q.from_z coeff)) Q.zero l in
  r;;

let eval_linear_atomic (eval_real_variable : int -> Q.t)
  (eval_prop_variable : string -> bool) : linear_atomic -> bool = function
    | `Propositional x -> eval_prop_variable x
    | `Greater_eq l ->
	Q.sgn (eval_linear_expression eval_real_variable l) >= 0;;

let eval_linear_formula (eval_real_variable : int -> Q.t)
  (eval_prop_variable : string -> bool) : linear_unquantified_formula -> bool =
  eval_formula (eval_linear_atomic eval_real_variable eval_prop_variable);;

let update_trans_var_rev_table (ct : linearization_context) =
  ct.trans_var_rev_table <- Array.create ct.trans_nr_vars "";
  Hashtbl.iter (fun name id ->
		  ct.trans_var_rev_table.(id) <- name) ct.trans_var_table;;

let process_unquantified_formula (f : source_unquantified_formula) :
    linearization_context * linear_unquantified_formula =
  let ct = empty_linearization_context () in
  let f' = linearize_unquantified_formula ct f in
    update_trans_var_rev_table ct;
    ct, f';;

module LinearExpressionSet = Set.Make
  (struct
     type t = linear_expression
     let compare : t -> t -> int = compare
   end);;

module LinearExpressionMap = Map.Make
  (struct
     type t = linear_expression
     let compare : t -> t -> int = compare
   end);;

let rec atomic_formula_iter (f : 'atomic) :
    ('atomic, 'variable) quantified_formula -> unit =
  function
    | `And l | `Or l | `Xor l -> List.iter (atomic_formula_iter f) l
    | `Atomic a -> f a
    | `Not x | `Exists(_, x) | `Forall(_, x) -> atomic_formula_iter f x
    | `Ite(c, f1, f2) ->
	(atomic_formula_iter f c;
         atomic_formula_iter f f1;
	 atomic_formula_iter f f2);;

let rec atomic_formula_fold (f : 'atomic -> 'a -> 'a)
    (formula : ('variable, 'atomic) quantified_formula) (a : 'a) =
  match formula with
    | `And l | `Or l | `Xor l -> List.fold_right (atomic_formula_fold f) l a
    | `Atomic w -> f w a
    | `Not x | `Exists(_, x) | `Forall(_, x) -> atomic_formula_fold f x a
    | `Ite(c, f1, f2) ->
        atomic_formula_fold f c
          (atomic_formula_fold f f1
            (atomic_formula_fold f f2 a));;

let rec nnf_atomic_formula_fold f
    (formula : 'atomic nnf_unquantified_formula) (a : 'a) : 'a =
  match formula with
    | `And l | `Or l -> List.fold_right (nnf_atomic_formula_fold f) l a
    | ((`Atomic _) | (`Not (`Atomic _))) as w -> f w a;;

let collect_linear_atomic_formulas_in_unquantified
    (formula : linear_unquantified_formula) : LinearExpressionSet.t =
  atomic_formula_fold
    (fun (atomic : linear_atomic) s ->
       match atomic with
	 | `Greater_eq l -> LinearExpressionSet.add l s
	   | _ -> s)
    (unquantified_to_quantified formula) LinearExpressionSet.empty;;

let rec collect_source_variables_in_expression (f : source_expression)
    (before : StringSet.t) = match f with
      | `Sum l | `Product l -> List.fold_right
	  collect_source_variables_in_expression l before
      | `Neg e ->  collect_source_variables_in_expression e before
      | `Quotient(e1, e2) ->
	  collect_source_variables_in_expression e1
	    (collect_source_variables_in_expression e2 before)
      | `Integer _ -> before
      | `Variable x -> StringSet.add x before;;

let collect_source_variables_in_atomic (f : source_atomic) 
    (before : StringSet.t) = match f with
      | `Propositional _ -> before
      | `Comparison(lhs, op, rhs) ->
	  collect_source_variables_in_expression lhs
	    (collect_source_variables_in_expression rhs before);;

let collect_source_variables_in_unquantified
    (formula : source_unquantified_formula) : StringSet.t =
  atomic_formula_fold collect_source_variables_in_atomic
    (unquantified_to_quantified formula) StringSet.empty;;

let simplify_unquantified_formula (simplify_atomic : bool -> 'a -> 'b unquantified_formula) =
  let rec simplify_or (already : 'b unquantified_formula list) = function
    | [] ->
	(match already with
	   | [h] -> h
	   | v -> `Or v)
    | h::t -> (match simplify_formula h with
		 | `And [] -> `And []
		 | `Or [] -> simplify_or already t
		 | h' -> simplify_or (h'::already) t)

  and simplify_and (already : 'b unquantified_formula list) = function
    | [] ->
	(match already with
	   | [h] -> h
	   | v -> `And v)
    | h::t -> (match simplify_formula h with
		 | `Or [] -> `Or []
		 | `And [] -> simplify_and already t
		 | h' -> simplify_and (h'::already) t)

  and simplify_formula :'a unquantified_formula -> 'a unquantified_formula =
  function
      | `And v -> simplify_and [] v
      | `Or v -> simplify_or [] v
      | `Not(`And v) -> simplify_or [] (List.map (fun x -> `Not x) v)
      | `Not(`Or v) -> simplify_and [] (List.map (fun x -> `Not x) v)
      | `Not(`Not v) -> simplify_formula v
      | `Xor v -> `Xor (List.map simplify_formula v)
      | `Ite(c, f1, f2) -> `Ite((simplify_formula c),
				(simplify_formula f1),
				(simplify_formula f2))
      | `Not(`Ite(c, f1, f2)) -> `Ite((simplify_formula c),
				   (simplify_formula (`Not f1)),
				   (simplify_formula (`Not f2)))
      | `Not(`Xor v) -> `Not(`Xor (List.map simplify_formula v))
      | `Atomic f -> simplify_atomic false f
      | `Not(`Atomic f) -> simplify_atomic true f

  in simplify_formula;;

let simplify_source_atomic (negate : bool)
    (atom : source_atomic) : source_atomic unquantified_formula =
  if not negate
  then `Atomic atom
  else
    match atom with
      | `Propositional x -> `Not(`Atomic(`Propositional x))
      | `Comparison(lhs, op, rhs) ->
	  let op' = match op with
	    | `LESS -> `GREATER_EQ
	    | `LESS_EQ -> `GREATER
	    | `GREATER -> `LESS_EQ
	    | `GREATER_EQ -> `LESS
	    | `EQUAL -> `UNEQUAL
	    | `UNEQUAL -> `EQUAL in
	  `Atomic(`Comparison(lhs, op', rhs));;

let simplify_unquantified_source_formula = simplify_unquantified_formula simplify_source_atomic;;

let rec subst_expression (f : string -> source_expression) :
    source_expression -> source_expression = function
      | `Sum l -> `Sum(List.map (subst_expression f) l)
      | `Product l -> `Product(List.map (subst_expression f) l)
      | `Quotient(x, y) -> `Quotient((subst_expression f x),
				     (subst_expression f y))
      | `Neg x -> `Neg(subst_expression f x)
      | (`Integer _) as w -> w
      | (`Variable v) as w -> try f v with Not_found -> w;;

let subst_atomic (f : string -> source_expression) :
    source_atomic -> source_atomic = function
      | (`Propositional _) as w -> w
      | `Comparison(lhs, op, rhs) ->
	  `Comparison((subst_expression f lhs), op,
		      (subst_expression f rhs));;

let rec subst_quantified_formula (f : string -> source_expression) :
    source_quantified_formula -> source_quantified_formula = function
      | `And l -> `And(List.map (subst_quantified_formula f) l)
      | `Or l -> `Or(List.map (subst_quantified_formula f) l)
      | `Xor l -> `Xor(List.map (subst_quantified_formula f) l)
      | `Not x -> `Not(subst_quantified_formula f x)
      | `Ite(c, f1, f2) -> `Ite((subst_quantified_formula f c),
				(subst_quantified_formula f f1),
				(subst_quantified_formula f f2))
      | `Atomic a -> `Atomic(subst_atomic f a)
      | `Forall(vars, g) ->
	  `Forall(vars, (subst_quantified_formula
              (fun x -> if List.mem x vars then raise Not_found else f x) g))
      | `Exists(vars, g) ->
	  `Exists(vars, (subst_quantified_formula
              (fun x -> if List.mem x vars then raise Not_found else f x) g));;

let rec subst_unquantified_formula (f : string -> source_expression) :
    source_unquantified_formula -> source_unquantified_formula = function
      | `And l -> `And(List.map (subst_unquantified_formula f) l)
      | `Or l -> `Or(List.map (subst_unquantified_formula f) l)
      | `Xor l -> `Xor(List.map (subst_unquantified_formula f) l)
      | `Not x -> `Not(subst_unquantified_formula f x)
      | `Ite(c, f1, f2) -> `Ite((subst_unquantified_formula f c),
				(subst_unquantified_formula f f1),
				(subst_unquantified_formula f f2))
      | `Atomic a -> `Atomic(subst_atomic f a);;

let rec linear_expression_get_dimension = function
  | [] -> 0
  | [v, _] -> v+1
  | (v1, _)::(((v2, _)::_) as t) ->
      (if v1 >= v2 then (Printf.printf "%d::%d\n" v1 v2; assert false));
      linear_expression_get_dimension t;;

let linear_constraints_get_dimension :
    (linear_expression * convex_comparison_op) list -> int =
  List.fold_left (fun w (l, _) -> max w (linear_expression_get_dimension l)) 0;;

let var_mapper_of_linearization_context ct id =
  ct.trans_var_rev_table.(id);;

let rec collect_free_variables_in_expression (bound : StringSet.t)
  (already : StringSet.t) : source_expression -> StringSet.t = function
  | `Sum l | `Product l ->
      List.fold_left(collect_free_variables_in_expression bound) already l
  | `Quotient(e1, e2) ->
      collect_free_variables_in_expression bound
	(collect_free_variables_in_expression bound already e1) e2
  | `Neg e1 -> collect_free_variables_in_expression bound already e1
  | `Variable x when not (StringSet.mem x bound) -> StringSet.add x already
  | `Integer _ | `Variable _ -> already

let collect_free_variables_in_atomic (bound : StringSet.t)
  (already : StringSet.t) = function
    | `Propositional _ -> already
    | `Comparison(e1, _, e2) ->
	collect_free_variables_in_expression bound
	  (collect_free_variables_in_expression bound already e1) e2
       
let rec collect_free_variables_in_formula (bound : StringSet.t)
  (already : StringSet.t) :
  source_quantified_formula -> StringSet.t = function
    | `Not f -> collect_free_variables_in_formula bound already f
    | `Atomic f -> collect_free_variables_in_atomic bound already f
    | ` And l | `Or l | `Xor l ->
       List.fold_left (collect_free_variables_in_formula bound) already l
    | `Ite(c, f1, f2) ->
       collect_free_variables_in_formula bound (collect_free_variables_in_formula bound
          (collect_free_variables_in_formula bound already c) f1) f2
    | `Forall(v, f) | `Exists(v, f) ->
	collect_free_variables_in_formula
          (add_list_to_string_set v bound)
	  already f;;

let existential_closure (f : source_quantified_formula) :
  source_quantified_formula =
  match StringSet.elements
    (collect_free_variables_in_formula StringSet.empty StringSet.empty f) with
    | [] -> f
    | v -> `Exists(v, f)

let stringset_of_list (l : string list) =
  List.fold_right StringSet.add l StringSet.empty;;

let existential_closure_minus (f : source_quantified_formula)
    (minus_variables : string list) :
  source_quantified_formula =
  match StringSet.elements
    (collect_free_variables_in_formula
       (stringset_of_list minus_variables) StringSet.empty f) with
    | [] -> f
    | v -> `Exists(v, f)

let universal_closure_minus (f : source_quantified_formula)
    (minus_variables : string list) :
  source_quantified_formula =
  match StringSet.elements
    (collect_free_variables_in_formula
       (stringset_of_list minus_variables) StringSet.empty f) with
    | [] -> f
    | v -> `Forall(v, f)

let output_syntax : (out_channel -> source_quantified_formula -> unit) Multiple_choices.choice_base =Multiple_choices.create "output-syntax" "Set output syntax";;
Mjollnir_options.add_option (Multiple_choices.arg_spec output_syntax);;
