{
open Formula_parser
open Parser_helper
}

let eol = "\r\n" | "\r" | "\n"
let whitespace = ['\r' '\t' ' ']+

rule normal_token = parse
| ['A'-'Z']  [ 'a'-'z' 'A'-'Z' '0'-'9' ]* {
  match Lexing.lexeme lexbuf with
  | "True" -> Lex_true((), (cur_extent lexbuf))
  | "False" -> Lex_false((), (cur_extent lexbuf))
  | "ForAll" | "Forall" -> Lex_forall((), (cur_extent lexbuf))
  | "Exists" -> Lex_exists((), (cur_extent lexbuf))

  | "Plus" -> Lex_kw_plus((), (cur_extent lexbuf))
  | "Neg"  -> Lex_kw_neg ((), (cur_extent lexbuf))
  | "Times" -> Lex_kw_times((), (cur_extent lexbuf))
  | "Divide" -> Lex_kw_divide((), (cur_extent lexbuf))

  | "And" -> Lex_kw_and((), (cur_extent lexbuf))
  | "Or"  -> Lex_kw_or ((), (cur_extent lexbuf))
  | "Xor" -> Lex_kw_xor((), (cur_extent lexbuf))
  | "Iff" -> Lex_kw_iff((), (cur_extent lexbuf))
  | "Not" -> Lex_kw_not((), (cur_extent lexbuf))
  | "Implies" -> Lex_kw_implies((), (cur_extent lexbuf))
  | "If" -> Lex_kw_if((), (cur_extent lexbuf))

  | id -> Lex_u_identifier(id, (cur_extent lexbuf)) }

| ['a'-'z']  [ 'a'-'z' 'A'-'Z' '0'-'9' ]* {
    Lex_l_identifier((Lexing.lexeme lexbuf), (cur_extent lexbuf)) }

| "(" { Lex_lparen((), (cur_extent lexbuf)) }
| ")" { Lex_rparen((), (cur_extent lexbuf)) }
| "[" { Lex_lbracket((), (cur_extent lexbuf)) }
| "]" { Lex_rbracket((), (cur_extent lexbuf)) }
| "{" { Lex_lbrace((), (cur_extent lexbuf)) }
| "}" { Lex_rbrace((), (cur_extent lexbuf)) }

| "&&" { Lex_and((), (cur_extent lexbuf)) }
| "||" { Lex_or((), (cur_extent lexbuf)) }
| "!" { Lex_not((), (cur_extent lexbuf)) }


| "+" { Lex_plus((), (cur_extent lexbuf)) }
| "-" { Lex_minus((), (cur_extent lexbuf)) }
| "*" { Lex_multiply((), (cur_extent lexbuf)) }
| "/" { Lex_divide((), (cur_extent lexbuf)) }

| "==" { Lex_equal((), (cur_extent lexbuf)) }
| "!=" { Lex_not_equal((), (cur_extent lexbuf)) }
| "<" { Lex_less((), (cur_extent lexbuf)) }
| "<=" { Lex_less_equal((), (cur_extent lexbuf)) }
| ">" { Lex_greater((), (cur_extent lexbuf)) }
| ">=" { Lex_greater_equal((), (cur_extent lexbuf)) }

| "," { Lex_comma((), (cur_extent lexbuf)) }

| ['0'-'9']+ { Lex_integer(Z.from_string (Lexing.lexeme lexbuf),
	   (cur_extent lexbuf)) }

| whitespace { normal_token lexbuf }
| eof { Lex_EOF((), (cur_extent lexbuf)) }
| eol { next_line lexbuf; normal_token lexbuf }
| "(*" { skip_comment lexbuf; normal_token lexbuf }

and skip_comment = parse
  "*)" { () }
| eol { next_line lexbuf; skip_comment lexbuf }
| _ { skip_comment lexbuf }
