%{
  open Formula
  open Parser_helper
%}

%start proposition_eof

%token <string * Parser_helper.extent> Lex_u_identifier
%token <string * Parser_helper.extent> Lex_l_identifier

%token <unit * Parser_helper.extent> Lex_lparen
%token <unit * Parser_helper.extent> Lex_rparen
%token <unit * Parser_helper.extent> Lex_lbracket
%token <unit * Parser_helper.extent> Lex_rbracket
%token <unit * Parser_helper.extent> Lex_lbrace
%token <unit * Parser_helper.extent> Lex_rbrace

%token <unit * Parser_helper.extent> Lex_exists
%token <unit * Parser_helper.extent> Lex_forall

%token <unit * Parser_helper.extent> Lex_and
%token <unit * Parser_helper.extent> Lex_or
%token <unit * Parser_helper.extent> Lex_comma
%token <unit * Parser_helper.extent> Lex_plus
%token <unit * Parser_helper.extent> Lex_minus
%token <unit * Parser_helper.extent> Lex_multiply
%token <unit * Parser_helper.extent> Lex_divide

%token <unit * Parser_helper.extent> Lex_kw_plus
%token <unit * Parser_helper.extent> Lex_kw_times
%token <unit * Parser_helper.extent> Lex_kw_divide
%token <unit * Parser_helper.extent> Lex_kw_neg

%token <unit * Parser_helper.extent> Lex_kw_and
%token <unit * Parser_helper.extent> Lex_kw_or
%token <unit * Parser_helper.extent> Lex_kw_xor
%token <unit * Parser_helper.extent> Lex_kw_iff
%token <unit * Parser_helper.extent> Lex_kw_not
%token <unit * Parser_helper.extent> Lex_kw_implies
%token <unit * Parser_helper.extent> Lex_kw_if

%token <Z.t * Parser_helper.extent> Lex_integer

%token <unit * Parser_helper.extent> Lex_less
%token <unit * Parser_helper.extent> Lex_greater
%token <unit * Parser_helper.extent> Lex_less_equal
%token <unit * Parser_helper.extent> Lex_greater_equal
%token <unit * Parser_helper.extent> Lex_equal
%token <unit * Parser_helper.extent> Lex_not_equal

%token <unit * Parser_helper.extent> Lex_not
%token <unit * Parser_helper.extent> Lex_implies
%token <unit * Parser_helper.extent> Lex_true
%token <unit * Parser_helper.extent> Lex_false
%token <unit * Parser_helper.extent> Lex_EOF

%type <Formula.source_quantified_formula * Parser_helper.extent> proposition_eof
%type <Formula.source_quantified_formula * Parser_helper.extent> proposition
%type <Formula.source_expression * Parser_helper.extent> expression
%type <Formula.source_expression list * Parser_helper.extent> expression_sum_list
%type <Formula.comparison_op * Parser_helper.extent> comparison_op

%%

proposition_eof:
  proposition Lex_EOF { $1 }

proposition:
| proposition_or_list { `Or(fst $1), (snd $1) }
| proposition1 { $1 }

proposition1:
| proposition_and_list { `And(fst $1), (snd $1) }
| proposition2 { $1 }

proposition2:
| Lex_lparen proposition Lex_rparen { (fst $2), (merge_extents $1 $3) }

| atomic { `Atomic(fst $1), (snd $1) }

| Lex_not proposition2 { `Not(fst $2), (merge_extents $1 $2) }
| Lex_kw_implies Lex_lbracket proposition Lex_comma proposition Lex_rbracket
      { (`Or[(`Not (fst $3)); (fst $5)]), (merge_extents $1 $6)}
| Lex_kw_and Lex_lbracket proposition_list Lex_rbracket { `And(fst $3), (merge_extents $1 $4) }
| Lex_kw_or Lex_lbracket proposition_list Lex_rbracket { `Or(fst $3), (merge_extents $1 $4) }
| Lex_kw_xor Lex_lbracket proposition_list Lex_rbracket { `Xor(fst $3), (merge_extents $1 $4) }
| Lex_kw_not Lex_lbracket proposition Lex_rbracket { `Not(fst $3), (merge_extents $1 $4) }
| Lex_forall Lex_lbracket variables Lex_comma proposition Lex_rbracket
      { (`Forall((fst $3), (fst $5))),(merge_extents $1 $6) }
| Lex_exists Lex_lbracket variables Lex_comma proposition Lex_rbracket
      { (`Exists((fst $3), (fst $5))),(merge_extents $1 $6) }
| Lex_kw_if Lex_lbracket proposition Lex_comma proposition Lex_comma proposition Lex_rbracket
      { (`Ite((fst $3), (fst $5), (fst $7))), (merge_extents $1 $8) }
| Lex_true { (`And[]), (snd $1) }
| Lex_false { (`Or[]), (snd $1) }

variables:
| Lex_l_identifier { [fst $1], (snd $1) }
| Lex_lbrace variable_list0 Lex_rbrace { $2, (merge_extents $1 $3) }

variable_list0:
| /* */ { [] }
| variable_list { $1 }

variable_list:
| Lex_l_identifier { [fst $1] }
| Lex_l_identifier Lex_comma variable_list { (fst $1)::$3 }

proposition_list:
| proposition { [fst $1], (snd $1) }
| proposition Lex_comma proposition_list {(fst $1)::(fst $3), (merge_extents $1 $3)}

proposition_or_list:
| proposition1 Lex_or proposition1 {[(fst $1); (fst $3)], (merge_extents $1 $3)}
| proposition1 Lex_or proposition_or_list {(fst $1)::(fst $3), (merge_extents $1 $3)}

proposition_and_list:
| proposition2 Lex_and proposition2 {[(fst $1); (fst $3)], (merge_extents $1 $3)}
| proposition2 Lex_and proposition_and_list {(fst $1)::(fst $3), (merge_extents $1 $3)}

expression:
| expression_sum_list { `Sum(List.rev (fst $1)), (snd $1) }
| expression1 { $1 }

expression_sum_list:
| expression_sum_list Lex_minus expression1 { (`Neg(fst $3)::(fst $1)), (merge_extents $1 $3) }
| expression_sum_list Lex_plus expression1 { ((fst $3)::(fst $1)), (merge_extents $1 $3) }
| expression1 Lex_minus expression1 { [`Neg(fst $3); fst $1], (merge_extents $1 $3) }
| expression1 Lex_plus expression1 { [(fst $3); (fst $1)], (merge_extents $1 $3) }

expression_product_list:
| expression_product_list expression3 { ((fst $2)::(fst $1)), (merge_extents $1 $2) }
| expression2 expression3 { [(fst $2); (fst $1)], (merge_extents $1 $2) }

| expression_product_list Lex_multiply expression2 { ((fst $3)::(fst $1)), (merge_extents $1 $3) }
| expression2 Lex_multiply expression2 { [(fst $3); (fst $1)], (merge_extents $1 $3) }

expression1:
| expression_product_list { `Product(List.rev (fst $1)), (snd $1) }
| expression1 Lex_divide expression2 { `Quotient((fst $1), (fst $3)), (merge_extents $1 $3) }
| expression2 { $1 }

expression2:
| expression3 { $1 }
| Lex_minus expression3 { `Neg(fst $2), (merge_extents $1 $2) }
| Lex_plus expression3 { (fst $2), (merge_extents $1 $2) }

expression3:
| Lex_l_identifier { `Variable(fst $1), (snd $1) }
| Lex_integer { `Integer(fst $1), (snd $1) }
| Lex_lparen expression Lex_rparen { (fst $2), (merge_extents $1 $3) }
| Lex_kw_plus Lex_lbracket expression_list Lex_rbracket
    { `Sum(fst $3), (merge_extents $1 $4) }
| Lex_kw_times Lex_lbracket expression_list Lex_rbracket
    { `Product(fst $3), (merge_extents $1 $4) }
| Lex_kw_neg Lex_lbracket expression Lex_rbracket
    { `Neg(fst $3), (merge_extents $1 $4) }
| Lex_kw_divide Lex_lbracket expression Lex_comma expression Lex_rbracket
    { `Quotient((fst $3), (fst $5)), (merge_extents $1 $6) }

expression_list:
| expression { [fst $1], (snd $1) }
| expression Lex_comma expression_list { (fst $1)::(fst $3),
					 (merge_extents $1 $3) }

comparison_op:
| Lex_less { `LESS, (snd $1) }
| Lex_less_equal { `LESS_EQ, (snd $1) }
| Lex_greater { `GREATER, (snd $1) }
| Lex_greater_equal { `GREATER_EQ, (snd $1) }
| Lex_equal { `EQUAL, (snd $1) }
| Lex_not_equal { `UNEQUAL, (snd $1) }

atomic:
| expression comparison_op expression { `Comparison((fst $1), (fst $2), (fst $3)), (merge_extents $1 $3) }
| Lex_u_identifier { `Propositional(fst $1), (snd $1) }
