open Formula
  
let rec print_redlog_source_expression (channel : out_channel) : source_expression -> unit = function
  | `Sum[] -> output_string channel "0"
  | `Product[] -> output_string channel "1"
  | `Sum(l) ->
      output_string channel "(";
      print_redlog_source_expression_list "+" channel l;
      output_string channel ")"
  | `Product(l) ->
      output_string channel "(";
      print_redlog_source_expression_list "*" channel l;
      output_string channel ")"
  | `Quotient(x, y) ->
      output_string channel "(";
      print_redlog_source_expression channel x;
      output_string channel " / ";
      print_redlog_source_expression channel y;
      output_string channel ")"
  | `Neg(x) -> output_string channel "(- ";
      print_redlog_source_expression channel x;
      output_string channel ")"
  | `Integer(x) -> output_string channel (Z.to_string x)
  | `Variable(x) -> output_string channel x

and print_redlog_source_expression_list (separator : string) (channel : out_channel) : source_expression list -> unit = function
  | [] -> ()
  | [h] -> print_redlog_source_expression channel h
  | h::t ->  print_redlog_source_expression channel h;
     output_string channel separator;
     print_redlog_source_expression_list separator channel t;;

let print_redlog_source_atomic channel : source_atomic -> unit = function
  | `Propositional x -> output_string channel x
  | `Comparison(lhs, op, rhs) ->
      print_redlog_source_expression channel lhs;
      output_string channel
	(match op with
	   | `LESS -> "<"
	   | `GREATER -> ">"
	   | `LESS_EQ -> "<="
           | `GREATER_EQ -> ">="
           | `EQUAL -> "="
           | `UNEQUAL -> "!=");
      print_redlog_source_expression channel rhs;;

let print_redlog_quantified_formula (print_redlog_atomic : out_channel -> 'atomic -> unit)
  (channel : out_channel) : (string, 'atomic) quantified_formula -> unit =

  let rec print_redlog_rec = function
    | `And l -> print_redlog_list_op "true" "and" l
    | `Or l -> print_redlog_list_op "false" "or" l
    | `Xor l -> print_redlog_list_op "false" "xor" l
    | `Not x -> output_string channel "not (";
	print_redlog_rec x;
	output_string channel ")"
    | `Ite(_, _, _) -> failwith "redlog cannot print ite"
    | `Atomic x -> print_redlog_atomic channel x
    | `Forall(vars, x) ->
	print_redlog_quantifier "all" x vars
    | `Exists(vars, x) ->
	print_redlog_quantifier "ex" x vars

  and print_redlog_list_op neutral op l = 
    output_string channel "(";
    print_redlog_list_op_rec neutral op l;
    output_string channel ")"

  and print_redlog_list_op_rec neutral op = function
    | [] -> output_string channel neutral
    | [h] -> print_redlog_rec h
    | h::t ->
	print_redlog_rec h;
	output_string channel " ";
	output_string channel op;
	output_string channel " ";
	print_redlog_list_op neutral op t

  and print_redlog_quantifier quantifier formula = function
    | [] -> print_redlog_rec formula
    | h::t -> output_string channel quantifier;
	output_string channel "(";
	output_string channel h;
	output_string channel ", ";
	print_redlog_quantifier quantifier formula t;
	output_string channel ")"

  in print_redlog_rec;;

let print_redlog_source_quantified_formula :
    out_channel -> source_quantified_formula -> unit =
  print_redlog_quantified_formula print_redlog_source_atomic;;

Multiple_choices.add_choice output_syntax "redlog" 5 print_redlog_source_quantified_formula "LIRA syntax";;
