{
open Lira_parser
open Parser_helper
}

let eol = "\r\n" | "\r" | "\n"
let whitespace = ['\r' '\t' ' ']+

rule normal_token = parse
| '-'? ['0'-'9']+ { Lex_integer(Z.from_string (Lexing.lexeme lexbuf),
	   (cur_extent lexbuf)) }

| ['a'-'z' 'A'-'Z' '_'] [ 'a'-'z' 'A'-'Z' '_' '0'-'9' ]* {
    let extent = cur_extent lexbuf
    and token = Lexing.lexeme lexbuf in
      match String.uppercase token with
	| "TRUE" -> Lex_true((), extent)
	| "FALSE" -> Lex_false((), extent)
	| "FORALL" -> Lex_forall((), extent)
	| "EXISTS" -> Lex_exists((), extent)
	| "AND" -> Lex_kw_and((), extent)
	| "OR" -> Lex_kw_or((), extent)
	| "XOR" -> Lex_kw_xor((), extent)
	| "NOT" -> Lex_kw_not((), extent)
	| "IMPLIES" -> Lex_kw_implies((), extent)
	| _ -> Lex_identifier(token, extent)
  }

| "(" { Lex_lparen((), (cur_extent lexbuf)) }
| ")" { Lex_rparen((), (cur_extent lexbuf)) }
| ":" { Lex_colon((), (cur_extent lexbuf)) }

| "+" { Lex_plus((), (cur_extent lexbuf)) }
| "-" { Lex_minus((), (cur_extent lexbuf)) }
| "*" { Lex_multiply((), (cur_extent lexbuf)) }
| "/" { Lex_divide((), (cur_extent lexbuf)) }

| "=" { Lex_equal((), (cur_extent lexbuf)) }
| "<>" { Lex_not_equal((), (cur_extent lexbuf)) }
| "<" { Lex_less((), (cur_extent lexbuf)) }
| "<=" { Lex_less_equal((), (cur_extent lexbuf)) }
| ">" { Lex_greater((), (cur_extent lexbuf)) }
| ">=" { Lex_greater_equal((), (cur_extent lexbuf)) }

| whitespace { normal_token lexbuf }
| eof { Lex_EOF((), (cur_extent lexbuf)) }
| eol { next_line lexbuf; normal_token lexbuf }
| "/*" { skip_comment lexbuf; normal_token lexbuf }

and skip_comment = parse
  "*/" { () }
| eol { next_line lexbuf; skip_comment lexbuf }
| _ { skip_comment lexbuf }
