%{
  open Formula
  open Parser_helper
%}

%type <Formula.source_quantified_formula * Parser_helper.extent> proposition_eof
%start proposition_eof

%token <string * Parser_helper.extent> Lex_identifier
%token <Z.t * Parser_helper.extent> Lex_integer

%token <unit * Parser_helper.extent> Lex_lparen
%token <unit * Parser_helper.extent> Lex_rparen
%token <unit * Parser_helper.extent> Lex_colon

%token <unit * Parser_helper.extent> Lex_exists
%token <unit * Parser_helper.extent> Lex_forall

%token <unit * Parser_helper.extent> Lex_kw_and
%token <unit * Parser_helper.extent> Lex_kw_or
%token <unit * Parser_helper.extent> Lex_kw_xor
%token <unit * Parser_helper.extent> Lex_kw_not
%token <unit * Parser_helper.extent> Lex_kw_implies

%token <unit * Parser_helper.extent> Lex_plus
%token <unit * Parser_helper.extent> Lex_minus
%token <unit * Parser_helper.extent> Lex_multiply
%token <unit * Parser_helper.extent> Lex_divide

%token <unit * Parser_helper.extent> Lex_less
%token <unit * Parser_helper.extent> Lex_greater
%token <unit * Parser_helper.extent> Lex_less_equal
%token <unit * Parser_helper.extent> Lex_greater_equal
%token <unit * Parser_helper.extent> Lex_equal
%token <unit * Parser_helper.extent> Lex_not_equal

%token <unit * Parser_helper.extent> Lex_true
%token <unit * Parser_helper.extent> Lex_false

%token <unit * Parser_helper.extent> Lex_EOF

%%

proposition_eof:
| proposition Lex_EOF { $1 }

proposition:
| proposition0 Lex_kw_implies proposition0 { (mk_or (mk_not (fst $1)) (fst $3)), (merge_extents $1 $3) }
| proposition0 { $1 }
proposition0:
| proposition Lex_kw_or proposition1 { (mk_or (fst $1) (fst $3)), (merge_extents $1 $3) }
| proposition1 { $1 }

proposition1:
| proposition1 Lex_kw_xor proposition2 { (mk_xor (fst $1) (fst $3)), (merge_extents $1 $3) }
| proposition2 { $1 }

proposition2:
| proposition2 Lex_kw_and proposition3 { (mk_and (fst $1) (fst $3)), (merge_extents $1 $3) }
| proposition3 { $1 }

proposition3:
| Lex_kw_not proposition3 { (mk_not (fst $2)), (merge_extents $1 $2) }
| Lex_exists Lex_lparen Lex_identifier Lex_colon proposition Lex_rparen {
    (mk_exists (fst $3) (fst $5)), (merge_extents $1 $6)
  }
| Lex_forall Lex_lparen Lex_identifier Lex_colon proposition Lex_rparen {
    (mk_forall (fst $3) (fst $5)), (merge_extents $1 $6)
  }
| Lex_lparen proposition Lex_rparen {
    $2
  }
| atomic { `Atomic(fst $1), (snd $1) }

expression:
| expression Lex_plus expression1 {
    (mk_sum (fst $1) (fst $3)), (merge_extents $1 $3) }
| expression Lex_minus expression1 {
    (mk_sum (fst $1) (mk_neg (fst $3))), (merge_extents $1 $3) }
| expression1 { $1 }
| error { failwith "error while parsing expression" }

expression1:
| expression1 Lex_multiply expression2 {
    (mk_product (fst $1) (fst $3)), (merge_extents $1 $3) }
| expression1 Lex_divide expression2 {
    (`Quotient((fst $1), (fst $3))), (merge_extents $1 $3) }
| expression2 { $1 }

expression2:
| expression3 { $1 }
| Lex_minus expression3 { (mk_neg (fst $2)), (merge_extents $1 $2) }

expression3:
| Lex_identifier { `Variable(fst $1), (snd $1) }
| Lex_integer { `Integer(fst $1), (snd $1) }
| Lex_lparen expression Lex_rparen { (fst $2), (merge_extents $1 $3) }

comparison_op:
| Lex_less { `LESS, (snd $1) }
| Lex_less_equal { `LESS_EQ, (snd $1) }
| Lex_greater { `GREATER, (snd $1) }
| Lex_greater_equal { `GREATER_EQ, (snd $1) }
| Lex_equal { `EQUAL, (snd $1) }
| Lex_not_equal { `UNEQUAL, (snd $1) }

atomic:
| expression comparison_op expression { `Comparison((fst $1), (fst $2), (fst $3)), (merge_extents $1 $3) }
| error { failwith "error while parsing atomic" };;
