(*
let rec split_rec l r = function
  | []  -> (l, r)
  | [h] -> ((h :: l), r)
  | h1 :: h2 :: t -> split_rec (h1 :: l) (h2 :: r) t;;
 
let split s0 = split_rec [] [] s0
*)

let rec split_at already others n =
  if n<= 0
  then (List.rev already), others
  else
    match others with
      | [] -> [], []
      | h::t -> split_at (h::already) t (n-1);;

let split l = split_at [] l ((List.length l) / 2);;

let rec minimize1 (test : 'a list -> bool) (already : 'a list) :
    'a list -> 'a list = function
      | [] -> already
      | h::t ->
	  if test (already @ t)
	  then minimize1 test already t
	  else minimize1 test (h::already) t;;

let rec minimal_rec predicate sup s0 =
  match s0 with
    | [_]|[] -> s0
    | _ ->
	let l0, r0 = split s0 in
         if predicate (sup @ l0)
         then minimal_rec predicate sup l0
         else if predicate (sup @ r0)
                then minimal_rec predicate sup r0
                else let l = minimal_rec predicate (sup @ r0) l0 in
                     let r = minimal_rec predicate (sup @ l) r0 in
                     l @ r;;

let minimize2 predicate lower upper =
  if predicate lower
  then lower
  else minimal_rec predicate lower upper;;

type minimize_choice = Sequential | Divide | No_minimize

let minimize_choices = Multiple_choices.create "minimize" "Set conjunction minimization algorithm";;
Multiple_choices.add_choice minimize_choices "none" 0 No_minimize "None";;
Multiple_choices.add_choice minimize_choices "sequential" 20 Sequential "Sequential minimization";;
Multiple_choices.add_choice minimize_choices "divide" 10 Divide "Divide and conquer minimization";;
Mjollnir_options.add_option (Multiple_choices.arg_spec minimize_choices);;

let minimize predicate lower upper =
  match Multiple_choices.get_choice minimize_choices with
    | No_minimize -> upper
    | Sequential -> minimize1 predicate lower upper
    | Divide -> minimize2 predicate lower upper;;

type 'comparison atom = Formula.linear_expression * 'comparison

let rec int_list_compare (a : int list) (b : int list) =
  match a, b with
    | [], [] -> 0
    | [], (_::_) -> -1
    | (_::_), [] -> 1
    | (ah::at), (bh::bt) ->
	if ah < bh
	then -1
	else if ah > bh
	then 1
	else int_list_compare at bt;;

let rem_initial_minus_one = function
  | (-1)::t -> t
  | l -> l;;

let compare_atoms_smaller_first ((a, _) : 'c atom) ((b, _) : 'c atom) =
  int_list_compare (rem_initial_minus_one (List.map fst a))
    (rem_initial_minus_one (List.map fst b))

let sort_atoms_smaller_first l =
  List.sort compare_atoms_smaller_first l

let compare_atoms_larger_first ((a, _) : 'c atom) ((b, _) : 'c atom) =
  - (int_list_compare (List.map fst a) (List.map fst b))

let sort_atoms_larger_first l =
  List.sort compare_atoms_larger_first l

let swap_elements a i j =
  let tmp = a.(i) in
    a.(i) <- a.(j);
    a.(j) <- tmp;;

let array_shuffle a =
  let n=Array.length a in
  for i=0 to n-2
  do
    swap_elements a i (i+(Random.int (n-i)))
  done;
  a;;

let list_shuffle l = Array.to_list (array_shuffle (Array.of_list l))

type sort_atom_choice = Sort_none | Sort_small_first | Sort_large_first | Sort_shuffle 
let sort_atoms_choices = Multiple_choices.create "sort-atoms" "Set atom sorting algorithm (for minimization)";;
Multiple_choices.add_choice sort_atoms_choices "none" 40 Sort_none "None";;
Multiple_choices.add_choice sort_atoms_choices "small-first" 30 Sort_small_first "Smaller first";;
Multiple_choices.add_choice sort_atoms_choices "large-first" 20 Sort_large_first "Sort atoms larger first";;
Multiple_choices.add_choice sort_atoms_choices "shuffle" 0 Sort_shuffle "Random shuffle";;
Mjollnir_options.add_option (Multiple_choices.arg_spec sort_atoms_choices);;

let generalize (predicate : 'a atom list -> bool)
    (lower : 'a atom list) (upper : 'a atom list) : 'a atom list =
  minimize predicate lower
    (match Multiple_choices.get_choice sort_atoms_choices with
       | Sort_none -> upper
       | Sort_small_first -> sort_atoms_smaller_first upper
       | Sort_large_first -> sort_atoms_larger_first upper
       | Sort_shuffle -> list_shuffle upper);;
  
