open Formula;;
Projection_apron.init();;

let modulo_theory = ref formula_true;;
let eclosure = ref false;;

let parse_lira channel : source_quantified_formula = fst (Lira_parser.proposition_eof Lira_lexer.normal_token (Lexing.from_channel channel));;
let parse_math channel : source_quantified_formula = fst (Formula_parser.proposition_eof Formula_lexer.normal_token (Lexing.from_channel channel));;

let parse_benchmark tokenizer channel =
  Smtlib.the_formula := None;
  Smtlib_parser.benchmark tokenizer channel;
  match ! Smtlib.the_formula with
    | None -> failwith "parse_benchmark"
    | Some f -> f;;

let parse_smtlib channel : source_quantified_formula = unquantified_to_quantified (parse_benchmark Smtlib_tokenizer.token (Lexing.from_channel channel));;

let input_syntax=Multiple_choices.create "input-syntax" "Set input syntax";;
Multiple_choices.add_choice input_syntax "math" 20 parse_math "(restricted) Mathematica syntax";;
Multiple_choices.add_choice input_syntax "lira" 10 parse_lira "LIRA syntax";;
Multiple_choices.add_choice input_syntax "smtlib" 5 parse_smtlib "SMT-LIB QF_LRA syntax";;
Mjollnir_options.add_option (Multiple_choices.arg_spec input_syntax);;

let action = Multiple_choices.create "action" "Set the action to perform";;
Multiple_choices.add_choice action "qelim-fr" 10
  (fun modulo_theory formula -> unquantified_to_quantified
     (simplify_unquantified_source_formula
       (Ferrante_rackoff.elim_quantifiers_source modulo_theory formula)))
  "Eliminate quantifiers using Ferrante and Rackoff's method";;
Multiple_choices.add_choice action "qelim-lw" 0
  (fun modulo_theory formula -> unquantified_to_quantified
     (simplify_unquantified_source_formula
       (Loos_weispfenning.elim_quantifiers_source modulo_theory formula)))
  "Eliminate quantifiers using Loos and Weispfenning's method";;
Multiple_choices.add_choice action "qelim-m" 20
  (fun modulo_theory formula -> unquantified_to_quantified
     (simplify_unquantified_source_formula
	(Formula_yices.reduce modulo_theory formula)))
  "Eliminate quantifiers using Monniaux's method";;
Multiple_choices.add_choice action "qelim-m2" 0
   (fun modulo_theory formula -> unquantified_to_quantified
     (Algo2_yices.eliminate modulo_theory formula))
  "No action: pass formula through and leave it identical";;
Multiple_choices.add_choice action "none" 0
  (fun modulo_theory formula -> formula)
  "No action: pass formula through and leave it identical";;
Mjollnir_options.add_option (Multiple_choices.arg_spec action);;

let test_method = ref false;;
Mjollnir_options.add_option ("--test", (Arg.Set test_method), "Test the implementation against qelim-m");;

let print_quantified_formula channel formula =
  Multiple_choices.get_choice output_syntax channel formula;;

let process_modulo_theory_filename filename =
  let channel = open_in filename in
    try
      modulo_theory := mk_v_and [!modulo_theory;
				 quantified_to_unquantified (Multiple_choices.get_choice input_syntax channel)]
    with Quantified ->
      begin
	Printf.fprintf stderr "Quantified expressions cannot be used in modulo_theory";
	exit 1
      end;;

let process_channel channel =
  let formula : source_quantified_formula =
    Multiple_choices.get_choice input_syntax channel in

  if !test_method then
    let simplified_formula = Multiple_choices.get_choice action !modulo_theory formula in
    let should_be_unsatisfiable =
      existential_closure (Formula.mk_xor formula simplified_formula) in
    let should_be_false =
	Formula_yices.elim_quantifiers_source !modulo_theory should_be_unsatisfiable in
      (if should_be_false <> `Or[]
       then (output_string stderr "TEST FAILED\n";
	     print_quantified_formula stderr simplified_formula;
	     output_string stderr "\n";
	     print_quantified_formula stderr should_be_unsatisfiable;
	     output_string stderr "\n";
	     print_quantified_formula stderr (Formula.unquantified_to_quantified should_be_false);
	     output_string stderr "\n";
	     exit 1))
  else
    let preprocessed_formula =
      if !eclosure
      then existential_closure formula
      else formula in
      
      let simplified_formula =
        Multiple_choices.get_choice action !modulo_theory preprocessed_formula in  
        print_quantified_formula stdout simplified_formula;
        print_newline ();;

let processed_filename = ref false
let process_filename filename =
  processed_filename := true;
  process_channel (open_in filename);;

let () = Arg.parse
  (("--modulo-theory",
    (Arg.String process_modulo_theory_filename),
    "Parse the formula in <filename> and compute modulo it")::
   ("--eclosure",
     (Arg.Set eclosure),
    "Perform existential closure first")::
   (Mjollnir_options.get_options ()))
  process_filename
  "usage: mjollnir [options]";
  if not !processed_filename
  then process_channel stdin;;

