#include <assert.h>
#include <gmp.h>

#define NDEBUG
#undef DEBUG
#undef GC_DEBUG

#if defined(GC_DEBUG) || defined(DEBUG)
#include <stdio.h>
#endif

#define CAML_NAME_SPACE
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/custom.h>
#include <caml/alloc.h>
#include <caml/fail.h>
/* #include <caml/camlidlruntime.h> */

extern void camlidl_mpz_ptr_ml2c(value, mpz_ptr *);
extern value camlidl_mpz_ptr_c2ml(mpz_ptr *);
extern void camlidl_mpq_ptr_ml2c(value, mpq_ptr *);
extern value camlidl_mpq_ptr_c2ml(mpq_ptr *);

#include <yices_c.h>

#define CAMLID(id) ml_yices_##id
#define CAMLCALL(id) value CAMLID(id)

#ifdef __GNUC__
#define field(x) .x =
#else
#define field(x)
#endif

static inline yices_context context_of_value(value r) {
  return * ((yices_context *) (Data_custom_val(r)));
}

static inline value value_of_expr(yices_expr e) {
  return (value) e;
}

static inline yices_expr expr_of_value(value e) {
  return (yices_expr) e;
}

static inline value value_of_type(yices_type e) {
  return (value) e;
}

static inline yices_type type_of_value(value e) {
  return (yices_type) e;
}

static inline value value_of_var_decl(yices_var_decl e) {
  return (value) e;
}

static inline yices_type var_decl_of_value(value e) {
  return (yices_var_decl) e;
}

static inline yices_model model_of_value(value e) {
  return (yices_model) e;
}

static inline value value_of_model(yices_model e) {
  return (value) e;
}

static inline value value_of_lbool(lbool v) {
  switch(v) {
  case l_true: return 1;
  case l_false: return 3;
  default: return 5;
  }
}

void CAMLID(context_finalize)(value r) {
  /* TODO */
  yices_del_context (context_of_value(r));
#ifdef GC_DEBUG
  fputs("Yices.context finalized\n", stderr);
#endif
}

struct custom_operations CAMLID(custom_context) = {
    field(identifier) "Yices.context",
    field(finalize) CAMLID(context_finalize),
    field(compare) NULL,
    field(hash) NULL,
    field(serialize) NULL,
    field(deserialize) NULL,
};

CAMLCALL(version)(value dummy) {
  CAMLparam0();
  CAMLreturn(caml_copy_string(yices_version()));
}

CAMLCALL(mk_context)(value dummy) {
  CAMLparam0();
  CAMLlocal1(r);
  r = caml_alloc_custom(& CAMLID(custom_context),
			       sizeof(yices_context), 0, 1);
  * ((yices_context *) (Data_custom_val(r))) = yices_mk_context();
  CAMLreturn(r);
}

CAMLCALL(reset)(value context) {
  CAMLparam1(context);
  yices_reset(context_of_value(context));
  CAMLreturn(Val_unit);
}

CAMLCALL(dump_context)(value context) {
  CAMLparam1(context);
  yices_dump_context(context_of_value(context));
  CAMLreturn(Val_unit);
}

CAMLCALL(mk_true)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_expr(yices_mk_true(context_of_value(context))));
}

CAMLCALL(mk_false)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_expr(yices_mk_false(context_of_value(context))));
}

CAMLCALL(assert_expr)(value context, value expr) {
  CAMLparam2(context, expr);
  yices_assert(context_of_value(context), expr_of_value(expr));
  CAMLreturn(Val_unit);
}

CAMLCALL(check)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_lbool(yices_check(context_of_value(context))));
}

CAMLCALL(push)(value context) {
  CAMLparam1(context);
  yices_push(context_of_value(context));
  CAMLreturn(Val_unit);
}

CAMLCALL(pop)(value context) {
  CAMLparam1(context);
  yices_pop(context_of_value(context));
  CAMLreturn(Val_unit);
}

CAMLCALL(mk_bool_var)(value context, value name) {
  CAMLparam2(context, name);
  CAMLreturn(value_of_expr(yices_mk_bool_var(context_of_value(context), String_val(name))));
}

CAMLCALL(mk_fresh_bool_var)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_expr(yices_mk_fresh_bool_var(context_of_value(context))));
}

unsigned CAMLID(list_length)(value list) {
  unsigned length = 0;
  while(list != 1) {
   list = Field(list, 1);
    length++;
  }
  return length;
}

void CAMLID(expr_list_to_array)(value list, yices_expr* t, unsigned n) {
  for(unsigned i=0; i<n; i++) {
    assert(list != 1);
    t[i] = expr_of_value(Field(list, 0));
    list = Field(list, 1);
  }
  assert(list == 1);
}

#define mk_list_operator(op) \
CAMLCALL(mk_##op)(value context, value list) { \
  CAMLparam2(context, list); \
  int n = CAMLID(list_length)(list); \
  yices_expr args[n]; \
  CAMLID(expr_list_to_array)(list, args, n); \
  CAMLreturn(value_of_expr(yices_mk_##op(context_of_value(context), args, n))); \
} \
CAMLCALL(mk_##op##_a)(value context, value array) { \
  CAMLparam2(context, array); \
  int n = Wosize_val(array);						    \
  CAMLreturn(value_of_expr(yices_mk_##op(context_of_value(context), \
      (yices_expr*) (&Field(array, 0)), n)));		\
}

mk_list_operator(and)
mk_list_operator(or)
mk_list_operator(sum)
mk_list_operator(sub)
mk_list_operator(mul)

#define mk_unop(op) \
CAMLCALL(mk_##op)(value context, value expr) { \
  CAMLparam2(context, expr); \
  CAMLreturn(value_of_expr(yices_mk_##op(context_of_value(context), expr_of_value(expr)))); \
}

mk_unop(not)

CAMLCALL(mk_ite)(value context, value cond, value expr1, value expr2) {
  CAMLparam4(context, cond, expr1, expr2);
  CAMLreturn(value_of_expr(yices_mk_ite(context_of_value(context), expr_of_value(cond), expr_of_value(expr1), expr_of_value(expr2))));
}

#define mk_binop(op) \
CAMLCALL(mk_##op)(value context, value a1, value a2) { \
  CAMLparam3(context, a1, a2); \
  CAMLreturn(value_of_expr(yices_mk_##op(context_of_value(context), \
    expr_of_value(a1), expr_of_value(a2)))); \
}

mk_binop(lt)
mk_binop(gt)
mk_binop(le)
mk_binop(ge)
mk_binop(eq)
mk_binop(diseq)

CAMLCALL(mk_type)(value context, value type_name) {
  CAMLparam2(context, type_name);
  CAMLreturn(value_of_type(yices_mk_type(context_of_value(context), String_val(type_name))));
}

CAMLCALL(mk_function_type)(value context, value domain, value range) {
  CAMLparam3(context, domain, range);
  CAMLreturn(value_of_type(yices_mk_function_type(
      context_of_value(context),
      (yices_type*) (&Field(domain, 0)),
      Wosize_val(domain), type_of_value(range))));
}

CAMLCALL(mk_bitvector_type)(value context, value size) {
  CAMLparam2(context, size);
  CAMLreturn(value_of_type(yices_mk_bitvector_type(context_of_value(context), Int_val(size))));
}

CAMLCALL(mk_var_decl)(value context, value name, value type) {
  CAMLparam3(context, name, type);
  CAMLreturn(value_of_var_decl(yices_mk_var_decl(context_of_value(context), String_val(name), type_of_value(type))));
}

CAMLCALL(mk_bool_var_decl)(value context, value name) {
  CAMLparam2(context, name);
  CAMLreturn(value_of_var_decl(yices_mk_bool_var_decl(context_of_value(context), String_val(name))));
}

CAMLCALL(mk_var_from_decl)(value context, value var_decl) {
  CAMLparam2(context, var_decl);
  CAMLreturn(value_of_expr(yices_mk_var_from_decl(context_of_value(context), var_decl_of_value(var_decl))));
}

CAMLCALL(mk_num)(value context, value n) {
  CAMLparam2(context, n);
  CAMLreturn(value_of_expr(yices_mk_num(context_of_value(context), Int_val(n))));
}

CAMLCALL(mk_num_from_string)(value context, value n) {
  CAMLparam2(context, n);
  CAMLreturn(value_of_expr(yices_mk_num_from_string(context_of_value(context), String_val(n))));
}

CAMLCALL(mk_num_from_mpz)(value context, value _v_n) {
  CAMLparam2(context, _v_n);
  mpz_ptr n;
  camlidl_mpz_ptr_ml2c(_v_n, &n);

#ifdef DEBUG
  gmp_printf("mk_num_from_mpz(_, %Zd)\n", n);
  fflush(stdout);
#endif
  CAMLreturn(value_of_expr(yices_mk_num_from_mpz(context_of_value(context), n)));
}

CAMLCALL(mk_num_from_mpq)(value context, value _v_n) {
  CAMLparam2(context, _v_n);
  mpq_ptr n;
  camlidl_mpq_ptr_ml2c(_v_n, &n);

#ifdef DEBUG
  gmp_printf("mk_num_from_mpq(_, %Qd)\n", n);
  fflush(stdout);
#endif
  CAMLreturn(value_of_expr(yices_mk_num_from_mpq(context_of_value(context), n)));
}

CAMLCALL(mk_bv_constant)(value context, value size, value x) {
  CAMLparam3(context, size, x);
  CAMLreturn(value_of_expr(yices_mk_bv_constant(context_of_value(context),
				  Int_val(size),
						Int_val(x))));
}

CAMLCALL(mk_bv_constant_from_array)(value context, value vector) {
  CAMLparam2(context, vector);
  const unsigned size = Wosize_val(vector);
  int bv[size];
  for(unsigned i=0; i<size; i++) {
    bv[i] = Bool_val(Field(vector, i));
  }
  CAMLreturn(value_of_expr(yices_mk_bv_constant_from_array(context_of_value(context), size, bv)));
}

mk_binop(bv_add)
mk_binop(bv_sub)
mk_binop(bv_mul)
mk_unop(bv_minus)
mk_binop(bv_concat)
mk_binop(bv_and)
mk_binop(bv_or)
mk_binop(bv_xor)
mk_unop(bv_not)

mk_binop(bv_lt)
mk_binop(bv_le)
mk_binop(bv_gt)
mk_binop(bv_ge)

mk_binop(bv_slt)
mk_binop(bv_sle)
mk_binop(bv_sgt)
mk_binop(bv_sge)

#define mk_bv_shiftop(op) \
CAMLCALL(mk_bv_##op)(value context, value a, value nbits) { \
  CAMLparam3(context, a, nbits); \
  CAMLreturn(value_of_expr(yices_mk_bv_##op(context_of_value(context), expr_of_value(a), Int_val(nbits)))); \
}

mk_bv_shiftop(sign_extend)
mk_bv_shiftop(shift_left0)
mk_bv_shiftop(shift_left1)
mk_bv_shiftop(shift_right0)
mk_bv_shiftop(shift_right1)

CAMLCALL(get_model)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_model(yices_get_model(context_of_value(context))));
}

CAMLCALL(get_value)(value model, value decl) {
  CAMLparam2(model, decl);
  CAMLreturn(value_of_lbool(yices_get_value(model_of_value(model), var_decl_of_value(decl))));
}

CAMLCALL(get_int_value)(value model, value decl) {
  CAMLparam2(model, decl);
  if (model == 0) {
    caml_failwith("get_int_value: null model");
  }
  long ret;
  if (! yices_get_int_value(model_of_value(model), var_decl_of_value(decl), &ret)) {
    caml_failwith("get_int_value: does not fit in long");
  }
  value ret2 = Val_int(ret);
  if (Int_val(ret2) != ret) {
    caml_failwith("get_int_value: too large for Caml int");
  }
  CAMLreturn(ret2);
}

CAMLCALL(get_mpq_value)(value model, value decl) {
  CAMLparam2(model, decl);
  if (model == 0) {
    caml_failwith("get_mpq_value: null model");
  }
  mpq_t r;
  mpq_ptr r_p = r;
  mpq_init(r);
  if (!yices_get_mpq_value(model_of_value(model), var_decl_of_value(decl), r)){
    mpq_clear(r);
    caml_failwith("get_mpq_value: not a rational");
  }
#ifdef DEBUG
  gmp_printf("get_mpq_value = %Qd\n", r);
#endif
  CAMLreturn(camlidl_mpq_ptr_c2ml(& r_p));
}

CAMLCALL(get_mpz_value)(value model, value decl) {
  CAMLparam2(model, decl);
  if (model == 0) {
    caml_failwith("get_mpz_value: null model");
  }
  mpz_t r;
  mpz_ptr r_p = r;
  mpz_init(r);
  if (!yices_get_mpz_value(model_of_value(model), var_decl_of_value(decl), r)){
    mpz_clear(r);
    caml_failwith("get_mpz_value: not an integer");
  }
#ifdef DEBUG
  gmp_printf("get_mpz_value = %Zd\n", r);
#endif
  CAMLreturn(camlidl_mpz_ptr_c2ml(& r_p));
}

CAMLCALL(evaluate_in_model)(value model, value expr) {
  CAMLparam2(model, expr);
  CAMLreturn(value_of_lbool(yices_evaluate_in_model(model_of_value(model), expr_of_value(expr))));
}

/* DM - for some reason this function is not documented or declared */
extern yices_expr
yices_mk_function_update(yices_context ctx, yices_expr f, yices_expr *args[],
                         unsigned num_args, yices_expr new_val); 

CAMLCALL(mk_function_update)(value context, value f,
			     value args, value new_val) {
  CAMLparam4(context, f, args, new_val);
  CAMLreturn(value_of_expr(yices_mk_function_update(
			   context_of_value(context), expr_of_value(f),
			   (yices_expr**) (& Field(args, 0)),
			   Wosize_val(args), expr_of_value(new_val))));
}

CAMLCALL(mk_app)(value context, value f, value args) {
  CAMLparam3(context, f, args);
  CAMLreturn(value_of_expr(yices_mk_app(context_of_value(context),
      expr_of_value(f),
      (yices_expr*) (& Field(args, 0)),
      Wosize_val(args))));
}

CAMLCALL(get_bitvector_value)(value model, value decl, value ret) {
  CAMLparam3(model, decl, ret);
  unsigned n = Wosize_val(ret);
  int bv[n];
  if (! yices_get_bitvector_value(model_of_value(model), var_decl_of_value(decl), n, bv)) {
    caml_failwith("get_bitvector_value");
  }
  for(unsigned i=0; i<n; i++) {
    Field(ret, i) = Val_bool(bv[i]);
  }
  CAMLreturn(Val_unit);
}

CAMLCALL(get_int_bitvector_value)(value model, value decl, value bits) {
  CAMLparam3(model, decl, bits);
  unsigned n = Int_val(bits);
  int bv[n], ret=0;
  if (! yices_get_bitvector_value(model_of_value(model), var_decl_of_value(decl), n, bv)) {
    caml_failwith("get_int_bitvector_value");
  }
  for(int i=n-1; i>=0; i--) {
    ret = (ret << 1) | (bv[i] ? 1 : 0);
  }
  CAMLreturn(Val_int(ret));
}
