exception Unavailable

type 'a choice_base = {
  name : string;
  doc : string;
  choices : (string, string * 'a) Hashtbl.t;
  mutable default : (int * string * 'a) option;
  mutable current_choice : 'a  option };;

let create base_name doc = {
  name = base_name;
  doc = doc;
  choices = Hashtbl.create 10;
  default = None;
  current_choice = None};;

let add_choice (base : 'a choice_base) (id : string) (priority : int) (value : 'a) (doc : string) =
  (if Hashtbl.mem base.choices id
   then failwith (Printf.sprintf
     "multiple identifier '%s' in multiple choice base '%s'" id base.name));
  (match base.default with
    | Some(old_priority, old_id, _) when old_priority > priority
	|| old_priority = priority && old_id > id -> ()
    | _ -> (base.default <- Some(priority, id, value);
	   base.current_choice <- Some value));
  Hashtbl.add base.choices id (doc, value);;

let set_choice (base : 'a choice_base) (id : string) =
  try base.current_choice <- Some (snd (Hashtbl.find base.choices id))
  with Not_found -> raise Unavailable;;

let get_choice (base : 'a choice_base) =
  match base.current_choice with
    | Some x -> x
    | None -> failwith
	(Printf.sprintf "no module available for '%s'" base.name);;

let arg_parser (base : 'a choice_base) (id : string) =
  try set_choice base id
  with Unavailable ->
    let buffer = Buffer.create 30 in
      Buffer.add_string buffer "Choice '";
      Buffer.add_string buffer id;
      Buffer.add_string buffer "' is unavailable for '";
      Buffer.add_string buffer base.name;
      Buffer.add_string buffer "'\nAvailable choices are:";
      Hashtbl.iter (fun id (doc, _) ->
		      Buffer.add_string buffer "\n";
		      Buffer.add_string buffer id;
		      Buffer.add_string buffer ": ";
		      Buffer.add_string buffer doc) base.choices;
      raise (Arg.Bad (Buffer.contents buffer));;
      
let arg_spec (base : 'a choice_base) :
    Arg.key * Arg.spec * Arg.doc =
  (("--" ^ base.name), Arg.String(arg_parser base), base.doc);;

