type optimization_mode =
  | MINIMIZATION
  | MAXIMIZATION;;

type constraint_kind =
  | LESS
  | LESS_EQ
  | GREATER
  | GREATER_EQ
  | EQUAL;;

type 'variable zlinear_form = ('variable * Z.t) list;;
type ('variable, 'constraint_kind) zlconstraint = 'variable zlinear_form * 'constraint_kind * Z.t;;

type linear_expr;;

type lp_problem_status =
  | UNFEASIBLE_LP_PROBLEM
  | UNBOUNDED_LP_PROBLEM
  | OPTIMIZED_LP_PROBLEM;;

external linear_expr_from_variable : int->linear_expr = "ml_ppl_linear_expr_from_variable";;
external linear_expr_println : linear_expr->unit = "ml_ppl_linear_expr_println";;
external linear_expr_scale : Z.t -> linear_expr -> linear_expr = "ml_ppl_linear_expr_scale";;
external linear_expr_from_scaled_variable : Z.t -> int -> linear_expr = "ml_ppl_linear_expr_from_scaled_variable";;
external linear_expr_from_z : Z.t -> linear_expr = "ml_ppl_linear_expr_from_z";;
external linear_expr_add : linear_expr -> linear_expr -> linear_expr = "ml_ppl_linear_expr_add";;

let linear_expr_zero = linear_expr_from_z Z.zero

let linear_expr_make l =
  List.fold_right (fun (v, c) already ->
    linear_expr_add (linear_expr_from_scaled_variable c v) already) l
    linear_expr_zero;;

type linear_constraint;;
external constraint_make : linear_expr -> constraint_kind -> linear_expr -> linear_constraint = "ml_ppl_constraint_make";;
external constraint_println : linear_constraint -> unit = "ml_ppl_constraint_println";;

type degenerate_element = UNIVERSE | EMPTY
type nnc_poly
external nnc_poly_create : int -> degenerate_element -> nnc_poly = "ml_ppl_nnc_poly_create";;
external nnc_poly_println : nnc_poly -> unit = "ml_ppl_nnc_poly_println";;

external nnc_poly_add_constraint : nnc_poly -> linear_constraint -> unit = "ml_ppl_nnc_poly_add_constraint";;
external nnc_poly_add_constraint_and_minimize : nnc_poly -> linear_constraint -> unit = "ml_ppl_nnc_poly_add_constraint_and_minimize";;
external nnc_poly_add_constraints : nnc_poly -> linear_constraint list -> unit = "ml_ppl_nnc_poly_add_constraints";;
external nnc_poly_add_constraints_and_minimize : nnc_poly -> linear_constraint list -> unit = "ml_ppl_nnc_poly_add_constraints_and_minimize";;

external nnc_poly_remove_higher_space_dimensions : nnc_poly -> int -> unit = "ml_ppl_nnc_poly_remove_higher_space_dimensions";;

external nnc_poly_constraints : nnc_poly -> linear_constraint list = "ml_ppl_nnc_poly_constraints";;
external nnc_poly_minimized_constraints : nnc_poly -> linear_constraint list = "ml_ppl_nnc_poly_minimized_constraints";;

external nnc_poly_is_empty : nnc_poly -> bool = "ml_ppl_nnc_poly_is_empty";;
external nnc_poly_is_universe : nnc_poly -> bool = "ml_ppl_nnc_poly_is_universe";;

external constraint_read : linear_constraint -> (int, constraint_kind) zlconstraint = "ml_ppl_constraint_read";;

type generator;;

external generator_is_line : generator -> bool = "ml_ppl_generator_is_line";;
external generator_is_ray : generator -> bool = "ml_ppl_generator_is_ray";;
external generator_is_point : generator -> bool = "ml_ppl_generator_is_point";;
external generator_is_closure_point : generator -> bool = "ml_ppl_generator_is_closure_point";;

external generator_println : generator -> unit = "ml_ppl_generator_println";;
external generator_read_coefficients : generator -> int zlinear_form = "ml_ppl_generator_read_coefficients";;
external generator_read_divisor : generator -> Z.t = "ml_ppl_generator_read_divisor";;

type lp_problem;;

external lp_problem_solve : lp_problem -> lp_problem_status = "ml_ppl_lp_problem_solve";;

external lp_problem_create : linear_constraint list -> linear_expr -> optimization_mode -> lp_problem = "ml_ppl_lp_problem_create";;

external lp_problem_optimal_value : lp_problem -> Z.t * Z.t = "ml_ppl_lp_problem_optimal_value";;

external lp_problem_feasible_point : lp_problem -> generator = "ml_ppl_lp_problem_feasible_point";;

external lp_problem_optimizing_point : lp_problem -> generator = "ml_ppl_lp_problem_optimizing_point";;
