(* A locus is a location in the source text. *)
type locus = {
    locus_file_name   : string;
    locus_file_line   : int;
    locus_file_column : int;
    locus_file_char   : int; };;

(* An extent is a pair of loci, giving the beginning and the end of
   the syntactic element. *)
type extent = {
   extent_beg : locus;
   extent_end : locus };;

let merge_extents : ('a * extent) -> ('b * extent) -> extent =
  fun (_, a) (_, b) ->
    { extent_beg = a.extent_beg;
      extent_end = b.extent_end }

(* How to find locations in source *)
let current_line_num       = ref 1
let current_line_start_pos = ref 0
let current_file_name      = ref "*UNKNOWN*"

let next_line lexbuf =
  incr current_line_num;
  current_line_start_pos := Lexing.lexeme_end lexbuf;;

let find_locus_in_current_line n =
  (*incr num_loci;*)
  { locus_file_name   = !current_file_name;
    locus_file_line   = !current_line_num;
    locus_file_column = n - !current_line_start_pos;
    locus_file_char   = n; }

let cur_extent lexbuf =
  { extent_beg = 
    (find_locus_in_current_line (Lexing.lexeme_start lexbuf));
    extent_end =
    (find_locus_in_current_line (Lexing.lexeme_end lexbuf)) }
     
