exception Empty_polyhedron;;

let minimize_constraints_elapsed = ref 0.
and project_minimize_constraints_elapsed = ref 0.

type minimize_in = (Formula.linear_expression *
	      [`GREATER_EQ|`GREATER|`LESS_EQ|`LESS |`EQUAL ]) list
and minimize_out = (Formula.linear_expression *
              [`GREATER_EQ|`GREATER|`EQUAL]) list

let print_minimize_out (var_mapper : int->string) (channel : out_channel)
    (constraints : minimize_out) =
  output_string channel "And[";
  List.iter (fun (l_expr, op) ->
	      Formula_math.print_linear_expression channel var_mapper l_expr;
	      output_string channel
		(match op with
		   | `GREATER_EQ -> " >= 0"
		   | `GREATER -> " > 0"
		   | `EQUAL -> " = 0")) constraints;
  output_string channel "]";;
				 
type algo_choice = {
  minimize_constraints : minimize_in -> minimize_out;
  project_minimize_constraints : minimize_in -> int -> minimize_out
};;

let choices : algo_choice Multiple_choices.choice_base =
  Multiple_choices.create "polyhedra" "choice of polyhedra library";;

let minimize_constraints (cts : minimize_in) : minimize_out =
  Profiling.chronometer
    minimize_constraints_elapsed
    (Multiple_choices.get_choice choices).minimize_constraints cts;;

let project_minimize_constraints : minimize_in -> int -> minimize_out =
  fun cts dimension ->
  Profiling.chronometer project_minimize_constraints_elapsed
    ((Multiple_choices.get_choice choices).project_minimize_constraints cts) 
    dimension;;

Mjollnir_options.add_option (Multiple_choices.arg_spec choices);;
(*
let project_minimize_constraints_indices
    (input : minimize_in) (projected : int list) : minimize_out =
  print_string "projected: [";
  List.iter (fun i -> print_int i; print_string " ") projected;
  print_endline "]";
  project_minimize_constraints input (List.length projected)
*)
