let float_rounding = true

let vand l = `And l
let vor l = `Or l
let vnot x = `Not x
let implies p1 p2 = `Or [`Not p1; p2]

let less x y = `Atomic(`Comparison(x, `LESS, y))
let greater x y = `Atomic(`Comparison(x, `GREATER, y))
let less_eq x y = `Atomic(`Comparison(x, `LESS_EQ, y))
let greater_eq x y = `Atomic(`Comparison(x, `GREATER_EQ, y))
let equal x y = `Atomic(`Comparison(x, `EQUAL, y))

let const x = `Quotient((`Integer (Q.get_num x)), (`Integer (Q.get_den x)))
let var x = `Variable x
let plus x y = `Sum [x; y]
let minus x y = `Sum [x; `Neg y]
let scale s x = `Product[const s; x]
let neg x = `Neg x

let forall vars x = `Forall(vars, x)
let exists vars x = `Exists(vars, x)

let ifte c e1 e2 = vor [ vand [c; e1]; vand [vnot c; e2]];;

let fresh_variable =
  let table = Hashtbl.create 100 in
  function name ->
    let count =
      try Hashtbl.find table name
      with Not_found -> 0 in
    Hashtbl.replace table name (count + 1);
    Printf.sprintf "%s$%d" name count;;

let subst_var var_name value f =
  Formula.subst_quantified_formula (fun v -> if var_name=v then value else raise Not_found) f;;

let is_minimum v prop =
  let v' = fresh_variable v in
  forall [v'] (vand [ prop ;
		       (implies (subst_var v (var v') prop) (less_eq (var v) (var v'))) ]);;

let eps_rel = Q.div Q.one (Q.from_z (Z.mul2exp Z.one 53))
let eps_abs = Q.div Q.one (Q.from_z (Z.mul2exp Z.one 1023))

let is_rounded rounded unrounded =
  if float_rounding
  then
    vor [vand [less_eq (var rounded) (const eps_abs);
	       greater_eq (var rounded) (const Q.zero);
	       less_eq unrounded (const eps_abs);
	       greater_eq unrounded (const Q.zero)];
	 
	 vand [greater_eq (var rounded) (const (Q.neg eps_abs));
	       less_eq (var rounded) (const Q.zero);
	       greater_eq unrounded (const (Q.neg eps_abs));
	       less_eq unrounded (const Q.zero)];
	 
       vand [greater_eq (var rounded) (const eps_abs);
	     greater_eq unrounded (const eps_abs);
	     less_eq (var rounded) (scale (Q.add Q.one eps_rel) unrounded);
	     greater_eq (var rounded) (scale (Q.sub Q.one eps_rel) unrounded)];
       
       vand [less_eq (var rounded) (const (Q.neg eps_abs));
	     less_eq unrounded (const (Q.neg eps_abs));
	     greater_eq (var rounded) (scale (Q.add Q.one eps_rel) unrounded);
	     less_eq (var rounded) (scale (Q.sub Q.one eps_rel) unrounded)]]
  else equal (var rounded) unrounded;;

let correct =
  vand [  less_eq (var "e1min") (var "e1max");
	  less_eq (var "e2min") (var "e2max");
	  less_eq (var "e3min") (var "e3max");
	  greater (var "e2min") (const Q.zero) ];;

let postcondition =
  vand [ less_eq (var "s1") (var "s1max");
	 greater_eq (var "s1") (var "s1min") ];;

let precondition =
   vand [ greater_eq (var "e1") (var "e1min");
	  greater_eq (var "e2") (var "e2min");
	  greater_eq (var "e3") (var "e3min");

	  less_eq (var "e1") (var "e1max");
	  less_eq (var "e2") (var "e2max");
	  less_eq (var "e3") (var "e3max");

          less_eq (var "olds1") (var "s1max");
	  greater_eq (var "olds1") (var "s1min") ];;

let step1 =
  vand [ is_rounded "delta1" (minus (var "e1") (var "olds1"));
	 ifte (less (var "delta1") (var "e2")) (equal (var "o1") (var "e1"))
	     (equal (var "o1") (plus (var "olds1") (var "e2"))) ];;

let step2 =
  vand [ is_rounded "delta2" (minus (var "o1") (var "olds1"));
	 ifte (greater (var "delta2") (neg (var "e2")))
	     (equal (var "o2") (var "o1"))
	     (equal (var "o2") (minus (var "o1") (var "e2"))) ];;

let step3 = vor [ equal (var "s1") (var "o2");
		  equal (var "s1") (var "e3") ];;

let nonempty =  less_eq (var "s1min") (var "s1max");;

let stable = vand [nonempty;
  forall [ "olds1"; "e1"; "e2"; "e3"; "o1"; "o2"; "delta1"; "delta2"; "s1" ]
    (implies ( vand [ precondition; step1; step2; step3 ]) postcondition )] ;;

let stable_s1max = forall [] (exists [ "s1min" ] stable);;
(*
let do_substs =
  List.fold_right

     (fun (var, value) formula -> pr_subst var value formula)
*)
let do_substs _ f = f

let instances1 =
  [ (*"e1min", (const (Q.from_int (-10000)));
    "e1max", (const (Q.from_int (10000)));
    "e2min", (const (Q.from_int 1));
    "e2max", (const (Q.from_int 3));
    "e3min", (const (Q.from_int (-5000)));
    "e3max", (const (Q.from_int (5000))) *) ];;

print_endline "Stable sets:";;

let stable_s1max_elim = Formula.simplify_unquantified_source_formula (Formula_yices.elim_quantifiers_source correct stable_s1max);;
Formula.print_source_unquantified_formula stdout stable_s1max_elim;;
print_newline ();;

Gc.full_major();;

print_endline "Best sets:";;

let to_dnf_step x = `Exists ([], x);;

let best_s1max_elim = Formula.simplify_unquantified_source_formula (Formula_yices.elim_quantifiers_source correct (to_dnf_step (is_minimum "s1max"
  (stable_s1max_elim : Formula.source_unquantified_formula :> Formula.source_quantified_formula))));;
Formula.print_source_quantified_formula stdout
  (best_s1max_elim : Formula.source_unquantified_formula :> Formula.source_quantified_formula);;
print_newline ();;

Printf.printf "minimize_constraints = %g\nproject_minimize_constraints = %g\ncheck = %g\n"
  !Projection.minimize_constraints_elapsed
  !Projection.project_minimize_constraints_elapsed
  !Yices.check_elapsed;;
