Arg.parse (Mjollnir_options.get_options ())
  (fun x -> ())
  "usage: rlim_opt [options]";;

let float_rounding = true

let vand l = `And l
let vor l = `Or l
let vnot x = `Not x
let implies p1 p2 = `Or [`Not p1; p2]

let less x y = `Atomic(`Comparison(x, `LESS, y))
let greater x y = `Atomic(`Comparison(x, `GREATER, y))
let less_eq x y = `Atomic(`Comparison(x, `LESS_EQ, y))
let greater_eq x y = `Atomic(`Comparison(x, `GREATER_EQ, y))
let equal x y = `Atomic(`Comparison(x, `EQUAL, y))

let const x = `Quotient((`Integer (Q.get_num x)), (`Integer (Q.get_den x)))
let var x = `Variable x
let plus x y = `Sum [x; y]
let minus x y = `Sum [x; `Neg y]
let scale s x = `Product[const s; x]
let neg x = `Neg x

let forall vars x = `Forall(vars, x)
let exists vars x = `Exists(vars, x)

let ifte c e1 e2 = vor [ vand [c; e1]; vand [vnot c; e2]];;


let fresh_variable =
  let table = Hashtbl.create 100 in
  function name ->
    let count =
      try Hashtbl.find table name
      with Not_found -> 0 in
    Hashtbl.replace table name (count + 1);
    Printf.sprintf "%s$%d" name count;;

let subst_var var_name value f =
  Formula.subst_quantified_formula (fun v -> if var_name=v then value else raise Not_found) f;;

let is_minimum v prop =
  let v' = fresh_variable v in
  forall [v'] (vand [ prop ;
		       (implies (subst_var v (var v') prop) (less_eq (var v) (var v'))) ]);;

let is_maximum v prop =
  let v' = fresh_variable v in
  forall [v'] (vand [ prop ;
		       (implies (subst_var v (var v') prop) (greater_eq (var v) (var v'))) ]);;

let rel_error = Q.div Q.one (Q.from_z (Z.mul2exp Z.one 52))
let min_normal = Q.div Q.one (Q.from_z (Z.mul2exp Z.one 1023))
let min_denormal = Q.div Q.one (Q.from_z (Z.mul2exp Z.one 1075))

let is_add_rounded_positive rounded unrounded =
  vor [vand [less_eq unrounded (const min_normal);
	     equal rounded unrounded];
       vand [greater unrounded (const min_normal);
	     less_eq (minus rounded unrounded) (scale rel_error unrounded);
	     less_eq (minus unrounded rounded) (scale rel_error unrounded) ]];;
 
let is_add_rounded rounded unrounded =
  if float_rounding
  then
    vor [vand [greater_eq unrounded (const Q.zero);
	       is_add_rounded_positive (var rounded) unrounded];
	 
	 vand [less unrounded (const Q.zero);
	       is_add_rounded_positive (neg (var rounded)) unrounded]]

  else equal (var rounded) unrounded;;

(*
let e1min = const (Q.from_int 0)
and e1max = const (Q.from_int 10)
and e2min = const (Q.from_int 1)
and e2max = const (Q.from_int 1)
let e3min = e1min and e3max = e1max
and s1min = const (Q.from_int (-1))
and s1max = const (Q.from_int 10)
*)

let e1min = var "e1min"
and e1max = var "e1max"
and e2min = var "e2min"
and e2max = var "e2max"
let e3min = var "e3min"
and e3max = var "e3max"
and s1min = var "s1min"
and s1max = var "s1max"

let postcondition =
  vand [ less_eq (var "s1") (s1max);
	 greater_eq (var "s1") (s1min) ];;

let precondition =
   vand [ less_eq (var "e1") (e1max);
	  greater_eq (var "e1") (e1min);
	  less_eq (var "e2") (e2max);
	  greater_eq (var "e2") (e2min);
	  less_eq (var "e3") (e3max);
	  greater_eq (var "e3") (e3min);
	  less_eq (var "olds1") (s1max);
	  greater_eq (var "olds1") (s1min) ];;

let step1 =
  vand [ is_add_rounded "delta1" (minus (var "e1") (var "olds1"));
	 ifte (less (var "delta1") (var "e2"))
             (equal (var "o1") (var "e1"))
	     (equal (var "o1") (plus (var "olds1") (var "e2"))) ];;

let step2 =
  vand [ is_add_rounded "delta2" (minus (var "o1") (var "olds1"));
	 ifte (greater (var "delta2") (neg (var "e2")))
	     (equal (var "o2") (var "o1"))
	     (equal (var "o2") (minus (var "olds1") (var "e2"))) ];;

let step3 = vor [ equal (var "s1") (var "o2");
		  equal (var "s1") (var "e3") ];;

let correct =
  vand [ less_eq (e1min) (e1max);
	 less_eq (e2min) (e2max);
	 less_eq (e3min) (e3max);
	 less_eq (s1min) (s1max);
	 greater (e2min) (const Q.zero) ];;

let modulo_theory = correct;;

let stable = (
  (forall [ "olds1"; "e1"; "e2"; "e3"; "o1"; "o2"; "delta1"; "delta2"; "s1" ]
    (vand [correct;
	   implies ( vand [ precondition; step1; step2; step3 ] )
	     postcondition ])));;
(*
let do_substs =
  List.fold_right

     (fun (var, value) formula -> pr_subst var value formula)
*)
let do_substs _ f = f

let instances1 =
  [ (*"e1min", (const (Q.from_int (-10000)));
    "e1max", (const (Q.from_int (10000)));
    "e2min", (const (Q.from_int 1));
    "e2max", (const (Q.from_int 3));
    "e3min", (const (Q.from_int (-5000)));
    "e3max", (const (Q.from_int (5000))) *) ];;

(*print_endline "Stable sets:";;*)

let stable_s1max = forall [] (exists [ "s1min" ] stable);;
let stable_s1max_elim = stable_s1max;;

let stable_s1min = forall [] (exists [ "s1max" ] stable);;
let stable_s1min_elim = stable_s1min;;

Gc.full_major();;

(*print_endline "Best sets:";;*)


let best_s1max_elim =  (is_minimum "s1max"
  (stable_s1max_elim));;
Formula_math.print_source_quantified_formula stdout (best_s1max_elim);;
print_newline();;

let best_s1min_elim =  (is_maximum "s1min"
  (stable_s1min_elim));;
Formula_math.print_source_quantified_formula stdout (best_s1min_elim);;
print_newline();;
