type sat_status = Sat | Unsat | Unknown;;
let the_formula = ref None
let the_status = ref Unknown

type term = Formula.source_expression
type formula = Formula.source_unquantified_formula

let term_alias_hash : (string, term) Hashtbl.t = Hashtbl.create 100;;
let formula_alias_hash : (string, formula) Hashtbl.t = Hashtbl.create 100;;

let set_benchmark_name name =
  Printf.printf "Benchmark name: %s\n" name;;
let set_benchmark_logic logic =
  Printf.printf "Logic name: %s\n" logic;;
let push_benchmark_assumption (assumption : formula) = failwith "assumption"

let set_benchmark_formula (f : formula) =
  the_formula := Some f;;
let set_benchmark_status (status : sat_status) =
  the_status := status

let mk_true : formula = Formula.formula_true
let mk_false : formula = Formula.formula_false

let rec mk_list_congruence1 binary_rel first already = function
  | [] -> already
  | h::t -> mk_list_congruence1 binary_rel first
      ((binary_rel first h)::already) t;;

let mk_list_congruence binary_rel = function
  | [] -> mk_true
  | h::t -> `And (mk_list_congruence1 binary_rel h [] t);;

let mk_not (x : formula) : formula = `Not x
let mk_implies (x : formula) (y : formula) : formula = failwith "mk_implies"
let mk_ite (c : formula) (x : formula) (y : formula) : formula =
  `Ite(c, x, y)
let mk_or_list (l : formula list) : formula = `Or l
let mk_and_list (l : formula list) : formula = `And l
let mk_xor_list (l : formula list) : formula = `Xor l
let mk_iff_list (l : formula list) : formula = mk_list_congruence
  (fun a b -> `Not(`Xor [a; b])) l

let mk_binary_eq x y : formula = `Atomic(`Comparison(x, `EQUAL, y))

let mk_list_eq (l : term list) : formula =
  mk_list_congruence mk_binary_eq l

let mk_distinct (l : term list) : formula = failwith "mk_distinct"

let mk_lt (a : term) (b : term) : formula =
  `Atomic(`Comparison(a, `LESS, b))
let mk_leq (a : term) (b : term) : formula =
  `Atomic(`Comparison(a, `LESS_EQ, b))
let mk_gt (a : term) (b : term) : formula =
  `Atomic(`Comparison(a, `GREATER, b))
let mk_geq (a : term) (b : term) : formula =
  `Atomic(`Comparison(a, `GREATER_EQ, b))

let remove_term_alias (alias : string) =
  Hashtbl.remove term_alias_hash alias;;
let remove_formula_alias (alias : string) =
  Hashtbl.remove formula_alias_hash alias;;
let add_term_alias (alias : string)(value : term) =
  Hashtbl.add term_alias_hash alias value;;
let add_formula_alias (alias : string) (value : formula) =
  Hashtbl.add formula_alias_hash alias value;;

let append_formula (l : formula list) (e : formula) : formula list =
  List.append l [e]
let unit_formula_list (h : formula) : formula list = [h];;

let append_term (l : term list) (e : term) : term list = List.append l [e]
let unit_term_list (h : term) : term list = [h];;

let mk_var (name : string) : term =
  try Hashtbl.find term_alias_hash name
  with Not_found -> `Variable name
let mk_const (const : string) : term = `Integer (Z.from_string const)
let mk_zero = `Integer Z.zero

let mk_propvar (name : string) : formula =
  try Hashtbl.find formula_alias_hash name
  with Not_found -> mk_gt (mk_var name) mk_zero

let mk_sum_list (terms : term list) : term = `Sum terms;;
let mk_diff_list (terms : term list) : term =
  match terms with
    | [] -> `Sum []
    | [x] -> x
    | h::t -> `Sum(h::(List.map (fun x->`Neg x) t));;

let mk_prod_list (terms : term list) : term = `Product terms;;
let mk_div (a : term) (b : term) : term = `Quotient(a, b);;

let mk_opposite (a : term) : term = `Neg a;;
let mk_cond (c : formula) (a : term) (b : term) : term = failwith "mk_cond";;

