type lbool = L_TRUE | L_FALSE | L_UNKNOWN;;

let string_of_lbool = function
  | L_TRUE -> "true"
  | L_FALSE -> "false"
  | L_UNKNOWN -> "unknown";;

exception Unknown
let bool_of_lbool = function
  | L_TRUE -> true
  | L_FALSE -> false
  | L_UNKNOWN -> raise Unknown;;

external version : unit -> string = "ml_yices_version";;

type context
external mk_context : unit -> context = "ml_yices_mk_context";;
external dump_context : context -> unit = "ml_yices_dump_context";;
external check : context -> lbool = "ml_yices_check";;
external reset : context -> unit = "ml_yices_reset"

external push : context -> unit = "ml_yices_push";;
external pop : context -> unit = "ml_yices_pop";;

type yices_type
type smt_type = yices_type
external mk_type : context -> string -> yices_type = "ml_yices_mk_type";;
external mk_function_type : context -> yices_type array -> yices_type -> yices_type = "ml_yices_mk_function_type";;
external mk_bitvector_type : context -> int -> yices_type = "ml_yices_mk_bitvector_type";;

type var_decl
type expr
external mk_var_decl : context -> string -> yices_type -> var_decl = "ml_yices_mk_var_decl";;
external mk_bool_var_decl : context -> string -> var_decl = "ml_yices_mk_bool_var_decl";;
external mk_var_from_decl : context -> var_decl -> expr = "ml_yices_mk_var_from_decl";;

let mk_real_var_decl context name =
  mk_var_decl context name (mk_type context "real");;

let mk_real_var context name =
  mk_var_from_decl context (mk_real_var_decl context name);;

external mk_true : context -> expr = "ml_yices_mk_true";;
external mk_false : context -> expr = "ml_yices_mk_false";;
external assert_expr : context -> expr -> unit = "ml_yices_assert_expr";;
external mk_bool_var : context -> string -> expr = "ml_yices_mk_bool_var";;
external mk_fresh_bool_var : context -> expr = "ml_yices_mk_fresh_bool_var";;

external mk_and : context -> expr list -> expr = "ml_yices_mk_and";;
external mk_or : context -> expr list -> expr = "ml_yices_mk_or";;
external mk_sum : context -> expr list -> expr = "ml_yices_mk_sum";;
external mk_sub : context -> expr list -> expr = "ml_yices_mk_sub";;
external mk_mul : context -> expr list -> expr = "ml_yices_mk_mul";;

external mk_and_a : context -> expr array -> expr = "ml_yices_mk_and_a";;
external mk_or_a : context -> expr array -> expr = "ml_yices_mk_or_a";;
external mk_sum_a : context -> expr array -> expr = "ml_yices_mk_sum_a";;
external mk_sub_a : context -> expr array -> expr = "ml_yices_mk_sub_a";;
external mk_mul_a : context -> expr array -> expr = "ml_yices_mk_mul_a";;

external mk_lt : context -> expr -> expr -> expr = "ml_yices_mk_lt";;
external mk_le : context -> expr -> expr -> expr = "ml_yices_mk_le";;
external mk_gt : context -> expr -> expr -> expr = "ml_yices_mk_gt";;
external mk_ge : context -> expr -> expr -> expr = "ml_yices_mk_ge";;
external mk_eq : context -> expr -> expr -> expr = "ml_yices_mk_eq";;
external mk_diseq : context -> expr -> expr -> expr = "ml_yices_mk_diseq";;
external mk_ite : context -> expr -> expr -> expr -> expr = "ml_yices_mk_ite";;
external mk_not : context -> expr -> expr = "ml_yices_mk_not";;

external mk_num : context -> int -> expr = "ml_yices_mk_num";;
external mk_num_from_string : context -> string -> expr = "ml_yices_mk_num_from_string";;
external mk_num_from_mpz : context -> Z.t -> expr = "ml_yices_mk_num_from_mpz";;external mk_num_from_mpq : context -> Q.t -> expr = "ml_yices_mk_num_from_mpq";;external mk_function_update : context -> expr -> expr array -> expr -> expr =
    "ml_yices_mk_function_update"
external mk_app : context -> expr -> expr array -> expr =
    "ml_yices_mk_app";;

external mk_bv_constant : context -> int -> int -> expr = "ml_yices_mk_bv_constant";;

external mk_bv_add : context -> expr -> expr -> expr = "ml_yices_mk_bv_add";;
external mk_bv_sub : context -> expr -> expr -> expr = "ml_yices_mk_bv_sub";;
external mk_bv_mul : context -> expr -> expr -> expr = "ml_yices_mk_bv_mul";;
external mk_bv_concat : context -> expr -> expr -> expr = "ml_yices_mk_bv_concat";;
external mk_bv_and : context -> expr -> expr -> expr = "ml_yices_mk_bv_and";;
external mk_bv_or : context -> expr -> expr -> expr = "ml_yices_mk_bv_or";;
external mk_bv_xor : context -> expr -> expr -> expr = "ml_yices_mk_bv_xor";;
external mk_bv_minus : context -> expr -> expr = "ml_yices_mk_bv_minus";;
external mk_bv_not : context -> expr -> expr = "ml_yices_mk_bv_not";;

external mk_bv_lt : context -> expr -> expr -> expr = "ml_yices_mk_bv_lt";;
external mk_bv_le : context -> expr -> expr -> expr = "ml_yices_mk_bv_le";;
external mk_bv_gt : context -> expr -> expr -> expr = "ml_yices_mk_bv_gt";;
external mk_bv_ge : context -> expr -> expr -> expr = "ml_yices_mk_bv_ge";;

external mk_bv_slt : context -> expr -> expr -> expr = "ml_yices_mk_bv_slt";;
external mk_bv_sle : context -> expr -> expr -> expr = "ml_yices_mk_bv_sle";;
external mk_bv_sgt : context -> expr -> expr -> expr = "ml_yices_mk_bv_sgt";;
external mk_bv_sge : context -> expr -> expr -> expr = "ml_yices_mk_bv_sge";;

external mk_bv_sign_extend : context -> expr -> int -> expr = "ml_yices_mk_bv_sign_extend";;
external mk_bv_shift_left0 : context -> expr -> int -> expr = "ml_yices_mk_bv_shift_left0";;
external mk_bv_shift_left1 : context -> expr -> int -> expr = "ml_yices_mk_bv_shift_left1";;
external mk_bv_shift_right0 : context -> expr -> int -> expr = "ml_yices_mk_bv_shift_right0";;
external mk_bv_shift_right1 : context -> expr -> int -> expr = "ml_yices_mk_bv_shift_right1";;

type model;;
external get_model : context -> model = "ml_yices_get_model";;
external get_value : model -> var_decl -> lbool = "ml_yices_get_value";;
external get_mpq_value : model -> var_decl -> Q.t = "ml_yices_get_mpq_value";; 
external get_mpz_value : model -> var_decl -> Z.t = "ml_yices_get_mpz_value"
external get_int_value : model -> var_decl -> int = "ml_yices_get_int_value"
external get_bitvector_value : model -> var_decl -> bool array -> unit = "ml_yices_get_bitvector_value"
external get_int_bitvector_value : model -> var_decl -> int -> int = "ml_yices_get_int_bitvector_value"
external evaluate_in_model : model -> expr -> lbool = "ml_yices_evaluate_in_model";;
