#!/usr/bin/perl -W
use strict;
my %programs = ( 'LIRA' => 0.02, 'Mjollnir' => 0.012,
                 'Mathematica' => 0.13, 'Redlog' => 0.210,
                 'Mjollnir --no-add-blocking-to-g' => 1,
                 'Mjollnir --no-block-projected-model' => 1,
                 'Proof-of-concept' => 1,
                 'Mjollnir --action qelim-fr' => 1,
                 'SMT_yices_eager' => 1);
my %totalTime;
my %problemsSolved;
my %timeouts;
my %oom;
my $result;
{
  my $program = 'undefined';
  my $filename;
  my $killed;
  while(my $line = <>) {
      chomp $line;
      if ($line =~ /^File: (.*)$/) {
	  $filename = $1;
	  undef $result;
      } elsif (defined($programs{$line})) {
	  $program = $line;
	  $killed = 0;
      } elsif ($line =~ /^True/ || $line =~ /\(sat\)/) {
	  if (defined($result) && $result!=1) {
	      print "meow for $filename\n";
	  }
	  $result=1;
      } elsif ($line =~ /^False/ || $line =~ /\(unsat\)/) {
	  if (defined($result) && $result!=0) {
	      print "meow for $filename\n";
	  }
	  $result=0;
      } elsif ($line =~ /Killed/) {
          $timeouts{$program}++;
          $killed=1;
      } elsif ($line =~ /Aborted/ || $line =~ /^Fatal/ || $line=~ /Out_of_memory/) {
          $oom{$program}++;
	  $killed=2;
      } elsif ($line =~ /^user\s+([0-9]+)m([0-9]+\.[0-9]*)s/ && !$killed) {
	  my $time = $1*60+$2;
	  # print "$filename\t$program\t$time\n";
	  $totalTime{$program} += $time;
	  $problemsSolved{$program}++;
      }
  }
}

foreach my $program(keys(%programs)) {
    $timeouts{$program}=0 if !defined($timeouts{$program});
    $oom{$program}=0 if !defined($oom{$program}); 
    if (defined($problemsSolved{$program})) {
      print "$program: ", $problemsSolved{$program}, " ", $totalTime{$program}, " ", $totalTime{$program}/$problemsSolved{$program}, " ", $timeouts{$program}, " ", $oom{$program}, "\n";
    }
}
